'''
Created on Mar 18, 2020

@author: rch

What are the use cases of this skeleton?

The services provided by this basic process structure include

 #Model

 ##Instantiate the model
   The model is an image of reality at a particular time instant t_k.
   Its STATE is defined by the primary variables and by state variables
   It provides the derived variables at the current time instance in form
   of properties.
   For hierarchically defined model that can incorporate several loops
   over different subdomains with variables defined only on a fractions
   of the model the properties are gathered from the definition of the 
   the subdomains (subdomain - material model). Such heterogeneous
   domain must define the mapping between the shared and local geometry and 
   representation discretization.

 ##Model interface
   The model provides the Residuum function `F` that is calculated
   for the current STATE, Further, the model provides also the
   derivatives with respect to the primary state variables dR_U.
   These two methods are used by the TStep to implement the predictor
   corrector scheme. The TStep interprets and controls the state
   for the purpose of time stepping within the TLoop

 ##Model specialization
   Model configurability. A particular model can be implemented
   In a few steps exploiting the subclassing and configurability of the
   of the model.
   
 #Controller
   Interaction with the model
   The evaluation process is controlled by the SimController, 
   It can launch the calculation within the current thread
   without interaction / monitoring of the current calculation
   or in a separate thread. The former option is useful for debugging.
 - Time line mixin is used to define the time range and time stepping
   parameters. Its changes are controlled by the TLoop.
   In a monitoring mode it is directly linked with the visual object
   time (VOT) in the plot sheet that defines the time step to be visualized.

 # Hist
   The model itself has no history. History is recorded during the
   time stepping TStep by the time loop TLoop. Variables to be recorded
   are specified in the model. The history is used for the calculation
   of cumulative properties based on the recorded variables and for the
   implementation of plot functions showing these variables
   
 ##Configuration
   Hist is used for recording of the model states during the time
   stepping. The variables to be recorded are specified in the Model 
   configuration using the `record' list. The recording can be 
   either minimalistic - specifying only the primary variables U and state 
   variables. Any other derived variable can be calculated 
   using the property functions. The Model is however implemented
   as a state machine depending on a value at a single time instant.
   Should the time dimension be implicitly allowed? The answer is not.
   This would bring about the possibility to include time-space problems  
   or regularized time steps as well in the algorithm. TEST it.
   
   
 # Why not timeline
Why not simply include the history in the model implicitly and 
providing an interface for TStep, TCut

Then, not only a single time step but a history slice 
would be possible to use in the calculation of the properties. 
This view would completely separate the notion of fundamental
state or history, the trial steps and the step to be updated.
This fits into the notions of dynamic calculations.

This model would then provide access to U_n which is a time 
series already accepted and U_k is the time sequence that is 
being changed during the current iteration. Upon an increment
TStep would append a mapping U_l = f(U_k) to U_n.

The history might be either 
set explicitly from previous calculation or calculated within
the current time loop. 

The nice thing about this approach would be the fact that 
there is no copy of the model in the Hist object. 
The role of the Hist object would be to map the data either
into the file structure or visualization back ends.

A point that would be on the negative side is the memory 
consumption for large finite element calculations. There,
the U_tn steps should be stored on the disk and only the U_tk
should be in memory. This makes the postprocessing difficult.

So the most general storage strategy seems to be
to preserve the current concept with one primary and state 
variables being recorded by default. Additional variables 
can be registered for recording given properties.

Within this scheme, also the field variables as large-sized
arrays can be conveniently stored in the file store back end.

This leads to the design that requires subclassing of Hist
for cumulative variables and for further plotting methods

'''

from ibvpy.view.ui import BMCSTreeNode

import numpy as np
import traits.api as tr

from .tline_mixin import TLineMixIn


class ITLoop(tr.Interface):
    pass


class IHist(tr.Interface):
    pass


class ITStep(tr.Interface):
    pass


class IModel(tr.Interface):
    pass


@tr.provides(ITStep)
class TStep(tr.HasStrictTraits):

    model = tr.Instance(IModel)
    sim = tr.DelegatesTo('model')

    t_n = tr.Float(0.0)
    t_n1 = tr.Float(0.0)

    def init_state(self):
        self.t_n = 0
        self.t_n1 = 0
        self.model.init_state()
        self.linalg_sys.register_constraints(0, 1)
        for var, state in self.model.S.items():
            state[...] = 0

    trial_state_changed = tr.Event

    R_dR_dU = tr.Property(depends_on='trial_state_changed, t_n1')

    @tr.cached_property
    def _get_R_dR_dU(self):
        R = self.R
        dR_dU = self.dR_dU
        return R, dR_dU

    R = tr.Property(depends_on='trial_state_changed, t_n1')

    @tr.cached_property
    def _get_R(self):
        bc = self.model.bc
        R = self.model.F - self.t_n1
        return R

    R_norm = tr.Property(depends_on='trial_state_changed, t_n1')

    @tr.cached_property
    def _get_R_norm(self):
        R = self.R
        return np.sqrt(np.einsum('...i,...i', R, R))

    dR_dU = tr.Property(depends_on='trial_state_changed')

    @tr.cached_property
    def _get_dR_dU(self):
        return self.model.d_F_U

    linalg_sys = tr.Instance(LinAlgSys, ())

    def make_iter(self):
        R, dR_dU = self.R_dR_dU
        self.linalg_sys.A = dR_dU
        self.linalg_sys.b_0 = R
        self.linalg_sys.apply_constraints(
            self.step_flag, self.t_n, self.t_n1
        )

        d_U = self.linalg_sys.solve()
        #d_U = -R / dR_dU
        self.model.U_k += d_U
        self.trial_state_changed = True

    def make_incr(self, t_n1):
        self.model.U_n[...] = self.model.U_k
        self.t_n1 = t_n1


@tr.provides(ITLoop)
class TLoop(tr.HasTraits):

    tstep = tr.Instance(ITStep)
    sim = tr.DelegatesTo('tstep')
    tline = tr.Property

    def _get_tline(self):
        return self.sim.tline

    k_max = tr.Int(100, enter_set=True, auto_set=False)

    acc = tr.Float(1e-4, enter_set=True, auto_set=False)

    verbose = tr.Bool(False, enter_set=True, auto_set=False)

    paused = tr.Bool(False)

    restart = tr.Bool(True)

    user_wants_abort = tr.Property

    def _get_user_wants_abort(self):
        return self.restart or self.paused

    def init(self):
        if self.paused:
            self.paused = False
        if self.restart:
            self.tline.val = self.tline.min
            self.tstep.init_state()
            self.restart = False

    def eval(self):
        t_n1 = self.tline.val
        t_max = self.tline.max
        dt = self.tline.step

        if self.verbose:
            print('t:', end='')

        while t_n1 <= (t_max + 1e-8):
            if self.verbose:
                print('\t%5.2f' % t_n1, end='')
            k = 0
            self.tstep.t_n1 = t_n1
            while (k < self.k_max) and (not self.user_wants_abort):
                R_norm = self.tstep.R_norm
                if R_norm < self.acc:
                    if self.verbose:
                        print('(%g), ' % k, end='\n')
                    break
                try:
                    self.tstep.make_iter()
                except RuntimeError as e:
                    raise(e)
                k += 1
            else:  # handle unfinished iteration loop
                if k >= self.k_max:  # maximum number of restarts exceeded
                    # no success abort the simulation
                    self.restart = True
                    print('')
                    raise StopIteration('Warning: '
                                        'convergence not reached in %g iterations' % k)
                else:  # reduce the step size
                    dt /= 2
                    continue

            # accept the time step and record the state in history
            self.tstep.make_incr(t_n1)
            # update the line - launches notifiers to subscribers
            self.tline.val = min(t_n1, self.tline.max)
            # set a new target time
            t_n1 += dt
            self.tstep.t_n1 = t_n1
        return


@tr.provides(IHist)
class Hist(tr.HasStrictTraits):

    model = tr.Instance(IModel)
    tstep = tr.DelegatesTo('model')

    def init_state(self):
        pass


class SimControler(BMCSTreeNode, TLineMixIn):

    model = tr.Instance(IModel)
    hist = tr.DelegatesTo('model')
    tstep = tr.DelegatesTo('model')
    tloop = tr.DelegatesTo('model')


@tr.provides(IModel)
class Model(BMCSTreeNode):
    '''Contains the primary unknowns variables U_k
    '''
    tstep_type = tr.Type(TStep)
    tloop_type = tr.Type(TLoop)
    hist_type = tr.Type(Hist)
    sim_type = tr.Type(SimControler)

    tstep = tr.Property(depends_on='tstep_type')

    @tr.cached_property
    def _get_tstep(self):
        return self.tstep_type(model=self)

    tloop = tr.Property(depends_on='tloop_type')

    @tr.cached_property
    def _get_tloop(self):
        return self.tloop_type(tstep=self.tstep)

    hist = tr.Property(depends_on='hist_type')

    @tr.cached_property
    def _get_hist(self):
        return self.hist_type(model=self)

    sim = tr.Property(depends_on='sim_type')

    @tr.cached_property
    def _get_sim(self):
        return self.sim_type(model=self)

    bc = tr.List(tr.Callable)

    U_shape = tr.Tuple(1,)

    def init_state(self):
        self.U_k = np.zeros(self.U_shape, dtype=np.float64)
        self.U_n = np.copy(self.U_n)
        self.hist.init_state()

    def get_plot_sheet(self):
        return

    U_k = tr.Array(np.float64, TRIAL_STATE=True)
    U_n = tr.Array(np.float64, FUND_STATE=True)

    S = tr.Dict(tr.Str, tr.Array(np.float64), STATE=True)

    F = tr.Property(depends_on='+TRIAL_STATE,+INPUT')

    @tr.cached_property
    def _get_F(self):
        raise NotImplemented

    d_F_U = tr.Property(depends_on='+TRIAL_STATE,+INPUT')

    @tr.cached_property
    def _get_d_F_U(self):
        raise NotImplemented


class BCModel(Model):
    bc = [lambda t: t]
    '''List of vector functions that implement the mapping 
    F_\alpha = bar{F}(t) 
    and U_\alpha = \bar{U}_\alpha(t). 
    '''


class SinusModel(Model):

    #=========================================================================
    # Model implementation - F and Fprime
    #=========================================================================

    def _get_F(self):
        return np.sin(self.U_k)

    def _get_d_F_U(self):
        return [np.cos(self.U_k)]

    #=========================================================================
    # Derived variables
    #=========================================================================
    G = tr.Property

    def _get_G(self):
        return np.arctan(self.U_k)


if __name__ == '__main__':
    m = SinusModel()
    m.tloop.init()
    m.tloop.eval()
    print(m.U_k)
    pass
    # Configuration options:
    # Boundary conditions
    #
    # bc - dictionary of array based boundary conditions
    #
    # Add a plotting functionality - in fact this example rephrases
    # the situation from the current application. Now show the recording
    # functionality.
    #
    # define the recording of variables
    # define plotting functions
    # define save use case
    # define load use case




from traits.api import \
    HasStrictTraits, WeakRef, Any, Property, provides
from .i_fe_grid_slice import IFENodeSlice


class ISliceProperty(HasStrictTraits):

    fe_grid = WeakRef()

    def __getitem__(self, idx):
        return FEGridNodeSlice(fe_grid=self.fe_grid, grid_slice=idx)


@provides(IFENodeSlice)
class FEGridNodeSlice(HasStrictTraits):
    '''General implementation of a slice within the FEGrid
    '''

    fe_grid = WeakRef('ibvpy.mesh.fe_grid.FEGrid')

    grid_slice = Any

    def __repr__(self):
        return repr(self.grid_slice)

    dof_nodes = Property

    def _get_dof_nodes(self):
        return self.fe_grid.dof_grid.cell_grid.point_idx_grid[self.grid_slice]

    dofs = Property

    def _get_dofs(self):
        return self.fe_grid.dof_grid.dofs_Ia[self.dof_nodes, :]

    dof_X = Property

    def _get_dof_X(self):
        return self.fe_grid.dof_grid.cell_grid.point_X_arr[self.dof_nodes, :]

    geo_nodes = Property

    def _get_geo_nodes(self):
        return self.fe_grid.geo_grid.cell_grid.point_idx_grid[self.grid_slice]

    geo_x = Property

    def _get_geo_x(self):
        return self.fe_grid.geo_grid.cell_grid.point_x_arr[self.geo_nodes, :]

    geo_X = Property

    def _get_geo_X(self):
        return self.fe_grid.geo_grid.cell_grid.point_X_arr[self.geo_nodes, :]


if __name__ == '__main__':

    from ibvpy.fets.fets2D import FETS2D4Q
    from ibvpy.mesh.fe_grid import FEGrid
    fets_sample = FETS2D4Q()

    fe_domain = FEGrid(coord_max=(2., 3., ),
                       shape=(2, 3),
                       inactive_elems=[3],
                       fets_eval=fets_sample)

    fe_slice = fe_domain[:, -1, :, -1]
    print('elems')
    print(fe_slice.elems)
    print('dof_nodes')
    print(fe_slice.dof_nodes)
    print('dofs')
    print(fe_slice.dofs)
    print('dof coords')
    print(fe_slice.dof_X)
    print('geo_nodes')
    print(fe_slice.geo_nodes)
    print('geo_X')
    print(fe_slice.geo_X)
    print('parametric points')
    print(fe_slice.geo_x)



'''
'''

from bmcs_utils.api import \
    Model, Int, Float, FloatRangeEditor, View, Item, EitherType
from scipy.interpolate import interp1d
import traits.api as tr
import numpy as np


class TimeFunction(Model):
    name = 'time function'
    t_max = Float(1.0, TIME=True)
    t_shift = Float(0.0, TIME=True)

    def __call__(self, arg):
        return self.time_function(arg)

    time_function = tr.Property(depends_on='state_changed')
    @tr.cached_property
    def _get_time_function(self):
        return self._generate_time_function()

    def update_plot(self, axes):
        t_range = np.linspace(0,self.t_max,500)
        f_range = self.time_function(t_range)
        axes.plot(t_range, f_range)
        axes.set_xlabel(r'$t$ [-]')
        axes.set_ylabel(r'$\theta$ [-]')

class TFMonotonic(TimeFunction):

    ipw_view = View(
        Item('t_max'),
        Item('range_factor')
    )

    range_factor = Float(10000)

    def _generate_time_function(self):
        # Define the monotonic range beyond the unit range
        # allow for interactive extensions. The fact
        d_history = np.array([0, 1*self.range_factor])
        t_arr = np.array([0, self.t_max*self.range_factor])
        return interp1d(t_arr, d_history, bounds_error=False, fill_value='extrapolate')

class TFBilinear(TimeFunction):
    loading_ratio = Float(0.5, TIME=True)
    time_ratio = Float(0.5, TIME=True)

    ipw_view = View(
        Item('t_max'),
        Item('loading_ratio', editor=FloatRangeEditor(low=0, high=1)),
        Item('time_ratio', editor=FloatRangeEditor(low=0, high=1)),
        Item('range_factor')
    )

    range_factor = Float(1)

    def _generate_time_function(self):
        # Define the monotonic range beyond the unit range
        # allow for interactive extensions. The fact
        d_history = np.array([0, 1 *  self.loading_ratio *self.range_factor , 1 *self.range_factor ])
        t_arr = np.array([0, self.t_max  * self.time_ratio * self.range_factor , self.t_max *self.range_factor])
        return interp1d(t_arr, d_history, bounds_error=False, fill_value='extrapolate')


class TFCyclicSymmetricIncreasing(TimeFunction):
    number_of_cycles = Int(10, TIME=True)

    ipw_view = View(
        Item('number_of_cycles'),
        Item('t_max')
    )

    def _generate_time_function(self):
        n_levels = int(self.number_of_cycles) * 2
        d_levels = np.linspace(0, 1, n_levels)
        d_levels.reshape(-1, 2)[:, 0] *= -1
        d_history = d_levels.flatten()
        t_arr = np.linspace(0, self.t_max, len(d_history))
        return interp1d(t_arr, d_history, bounds_error=False, fill_value=self.t_max)


class TFCyclicNonsymmetricIncreasing(TimeFunction):
    number_of_cycles = Int(10, TIME=True)

    ipw_view = View(
        Item('number_of_cycles'),
        Item('t_max')
    )

    def _generate_time_function(self):
        d_levels = np.linspace(0, 1, self.number_of_cycles * 2)
        d_levels.reshape(-1, 2)[:, 0] *= 0
        d_history = d_levels.flatten()
        t_arr = np.linspace(0, self.t_max, len(d_history))
        return interp1d(t_arr, d_history, bounds_error=False, fill_value=self.t_max)


class TFCyclicSymmetricConstant(TimeFunction):
    number_of_cycles = Int(10, TIME=True)

    ipw_view = View(
        Item('number_of_cycles'),
        Item('t_max')
    )

    def _generate_time_function(self):
        d_levels = np.zeros((self.number_of_cycles * 2, ))
        d_levels.reshape(-1, 2)[:, 0] = -1
        d_levels.reshape(-1, 2)[:, 1] = 1
        d_history = d_levels.flatten()
        t_arr = np.linspace(0, self.t_max, len(d_history))
        return interp1d(t_arr, d_history, bounds_error=False, fill_value=self.t_max)

import sympy as sp

class TFCyclicNonsymmetricConstant(TimeFunction):
    number_of_cycles = Int(10, TIME=True)
    unloading_ratio = Float(0.5, TIME=True)
    shift_cycles = Int(0, TIME=True)

    ipw_view = View(
        Item('number_of_cycles'),
        Item('shift_cycles'),
        Item('unloading_ratio', editor=FloatRangeEditor(low=0, high=1)),
        Item('t_max')
    )

    def _generate_time_function(self):
        d_1 = np.zeros(1)
        d_2 = np.zeros(((self.number_of_cycles + self.shift_cycles) * 2, ))
        d_2.reshape(-1, 2)[self.shift_cycles:, 0] = 1
        d_2.reshape(-1, 2)[self.shift_cycles:, 1] = self.unloading_ratio
        d_history = d_2.flatten()
        d_arr = np.hstack((d_1, d_history))
        t_arr = np.linspace(0, self.t_max, len(d_arr))
        return interp1d(t_arr, d_arr, bounds_error=False, fill_value=d_arr[-1])


class TFCyclicSin(TimeFunction):
    number_of_cycles = Int(10, TIME=True)
    phase_shift = Float(0, TIME=True)

    ipw_view = View(
        Item('number_of_cycles'),
        Item('phase_shift'),
        Item('t_max')
    )

    def _generate_time_function(self):
        t = sp.symbols(r't')
        p = self.phase_shift
        tf = sp.sin(2 * self.number_of_cycles * sp.pi * t / self.t_max)
        if p > 0:
            T = self.t_max / self.number_of_cycles
            pT = p * T
            tf = sp.Piecewise(
                (0, t < pT),
                (tf.subs(t, t-pT), True)
            )
        return sp.lambdify(t, tf, 'numpy')

class TFSelector(TimeFunction):
    name='time function'
    profile = EitherType(
        options=[
            ('monotonic', TFMonotonic),
            ('bilinear', TFBilinear),
            ('cyclic-sym-const', TFCyclicSymmetricConstant),
            ('cyclic-sym-incr', TFCyclicSymmetricIncreasing),
            ('cyclic-nonsym-const', TFCyclicNonsymmetricConstant),
            ('cyclic-nonsym-incr', TFCyclicNonsymmetricIncreasing)
        ],
        TIME=True
    )

    depends_on = ['profile']

    t_max = tr.Property(Float)
    def _get_t_max(self):
        return self.profile_.t_max

    ipw_view=View(
        Item('profile')
    )

    def __call__(self, arg):
        return self.profile_(arg)

    def update_plot(self, axes):
        print('xxxxx')
        self.profile_.update_plot(axes)


'''
Created on Feb 11, 2018

@author: rch
'''

import os

from tvtk.api import tvtk
from ibvpy.view.plot3d.viz3d import Vis3D, Viz3D

import numpy as np
import traits.api as tr


class Vis3DField(Vis3D):

    var = tr.Str('<unnamed>')

    def setup(self):
        self.new_dir()
        # make a loop over the XDomainModel
        #sim = self.sim
        ts = self.tstep
        fe_domain = ts.fe_domain
        vtk_point_list = []
        vtk_cell_list = []
        vtk_cell_offset_list = []
        vtk_cell_types_list = []
        point_offset = 0
        cell_offset = 0
        for domain in fe_domain:
            xdomain = domain.xmodel
            var_function = domain.tmodel.var_dict.get(self.var, None)
            if var_function == None or xdomain.hidden:
                continue
            fets = xdomain.fets
            DELTA_x_ab = fets.vtk_expand_operator
            n_c = fets.n_nodal_dofs
            vtk_points = np.einsum(
                'Ia,ab->Ib',
                xdomain.x_Eia.reshape(-1, n_c), DELTA_x_ab
            )
            vtk_point_list.append(vtk_points)
            cells, cell_offsets, cell_types = xdomain.get_vtk_cell_data(
                'nodes', point_offset, cell_offset)
            point_offset += vtk_points.shape[0]
            cell_offset += cells.shape[0]
            vtk_cell_list.append(cells)
            vtk_cell_offset_list.append(cell_offsets)
            vtk_cell_types_list.append(cell_types)
        if len(vtk_cell_types_list) == 0:
            raise ValueError(
                'Empty output for field variable %s in model %s' %
                (self.var, domain.tmodel)
            )
        vtk_cell_types = np.hstack(vtk_cell_types_list)
        vtk_cell_offsets = np.hstack(vtk_cell_offset_list)
        vtk_cells = np.hstack(vtk_cell_list)
        n_cells = vtk_cell_types.shape[0]
        vtk_cell_array = tvtk.CellArray()
        vtk_cell_array.set_cells(n_cells, vtk_cells)
        self.ug = tvtk.UnstructuredGrid(points=np.vstack(vtk_point_list))
        self.ug.set_cells(vtk_cell_types,
                          vtk_cell_offsets,
                          vtk_cell_array)


class Viz3DField(Viz3D):

    visible = tr.Property(tr.Bool)

    def _set_visible(self, visible):
        self.d.visible = visible

    def _get_visible(self):
        return self.d.visible

    def plot(self, vot):
        # self.vis3d.sim
        ts = self.vis3d.tstep
        idx = ts.hist.get_time_idx(vot)
        self.d.file_list = self.vis3d.file_list
        self.d.timestep = idx




import sys

from ibvpy.dots.dots_grid_eval import DOTSGridEval
from ibvpy.fets import FETSEval
from scipy.linalg import \
    inv
from traits.api import \
    Array, Float, Int, Property, cached_property

import numpy as np


print(sys.path)


#-------------------------------------------------------------------------
# FETS2D4Q - 4 nodes iso-parametric quadrilateral element (2D, linear, Lagrange family)
#-------------------------------------------------------------------------

#-------------------------------------------------------------------------
# Element Information:
#-------------------------------------------------------------------------
#
# Here an isoparametric element formulation is applied.
# The implemented shape functions are derived based on the
# ordering of the nodes of the parent element defined in
# '_node_coord_map' (see below)
#
#-------------------------------------------------------------------------


class FETS2D4Q(FETSEval):

    debug_on = True

    dots_class = DOTSGridEval

    # Dimensional mapping
    dim_slice = slice(0, 2)

    # Order of node positions for the formulation of shape function
    #
    dof_r = Array(value=[[-1, -1], [1, -1], [1, 1], [-1, 1]])
    geo_r = Array(value=[[-1, -1], [1, -1], [1, 1], [-1, 1]])

    n_e_dofs = Int(8,
                   FE=True,
                   enter_set=False, auto_set=True,
                   label='number of elemnet dofs'
                   )
    t = Float(1.0,
              CS=True,
              enter_set=False, auto_set=True,
              label='thickness')

    # Integration parameters
    #
    ngp_r = Int(2,
                FE=True,
                enter_set=False, auto_set=True,
                )
    ngp_s = Int(2, FE=True,
                enter_set=False, auto_set=True,
                )

    # Corner nodes are used for visualization
    vtk_r = Array(value=[[-1., -1.], [1., -1.], [1., 1.], [-1., 1.]])
    vtk_cells = [[0, 1, 2, 3]]
    vtk_cell_types = 'Quad'

    #vtk_point_ip_map = [0,1,3,2]
    n_nodal_dofs = Int(2)

    A_C = Property(depends_on='A_m,A_f')

    @cached_property
    def _get_A_C(self):
        return np.array((1.0, 1.0, 1.0), dtype=np.float64)

    #---------------------------------------------------------------------
    # Method required to represent the element geometry
    #---------------------------------------------------------------------
    def get_N_geo_mtx(self, r_pnt):
        '''
        Return the value of shape functions for the specified local coordinate r
        '''
        cx = np.array(self.geo_r, dtype='float64')
        Nr = np.array([[1 / 4. * (1 + r_pnt[0] * cx[i, 0]) * (1 + r_pnt[1] * cx[i, 1])
                        for i in range(0, 4)]])
        return Nr

    def get_dNr_geo_mtx(self, r_pnt):
        '''
        Return the matrix of shape function derivatives.
        Used for the conrcution of the Jacobi matrix.

        @TODO - the B matrix is used
        just for uniaxial bar here with a trivial differential
        operator.
        '''
        #cx = self._node_coord_map
        cx = np.array(self.geo_r, dtype='float64')
        dNr_geo = np.array([[1 / 4. * cx[i, 0] * (1 + r_pnt[1] * cx[i, 1])
                             for i in range(0, 4)],
                            [1 / 4. * cx[i, 1] * (1 + r_pnt[0] * cx[i, 0])
                             for i in range(0, 4)]])
        return dNr_geo

    #---------------------------------------------------------------------
    # Method delivering the shape functions for the field variables and their derivatives
    #---------------------------------------------------------------------
    def get_N_mtx(self, r_pnt):
        '''
        Returns the matrix of the shape functions used for the field approximation
        containing zero entries. The number of rows corresponds to the number of nodal
        dofs. The matrix is evaluated for the specified local coordinate r.
        '''
        Nr_geo = self.get_N_geo_mtx(r_pnt)
        I_mtx = np.identity(self.n_nodal_dofs, float)
        N_mtx_list = [I_mtx * Nr_geo[0, i] for i in range(0, Nr_geo.shape[1])]
        N_mtx = np.hstack(N_mtx_list)
        return N_mtx

    def get_dNr_mtx(self, r_pnt):
        '''
        Return the derivatives of the shape functions
        '''
        return self.get_dNr_geo_mtx(r_pnt)

    def get_B_mtx(self, r_pnt, X_mtx):
        J_mtx = self.get_J_mtx(r_pnt, X_mtx)
        dNr_mtx = self.get_dNr_mtx(r_pnt)
        dNx_mtx = np.dot(inv(J_mtx), dNr_mtx)
        Bx_mtx = np.zeros((3, 8), dtype='float64')
        for i in range(0, 4):
            Bx_mtx[0, i * 2] = dNx_mtx[0, i]
            Bx_mtx[1, i * 2 + 1] = dNx_mtx[1, i]
            Bx_mtx[2, i * 2] = dNx_mtx[1, i]
            Bx_mtx[2, i * 2 + 1] = dNx_mtx[0, i]
        return Bx_mtx


#----------------------- example --------------------

def example_with_new_domain():
    from ibvpy.api import \
        TStepper as TS, RTDofGraph, RTraceDomainListField, \
        RTraceDomainListInteg, TLoop, \
        TLine, BCDof, DOTSEval
    from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic
    from ibvpy.tmodel.mats2D.mats2D_sdamage.mats2D_sdamage import MATS2DScalarDamage

    from ibvpy.api import BCDofGroup

    mats_eval = MATS2DElastic()
    fets_eval = FETS2D4Q(mats_eval=mats_eval)
    #fets_eval = FETS2D4Q(mats_eval = MATS2DScalarDamage())

    print(fets_eval.vtk_node_cell_data)

    from ibvpy.mesh.fe_grid import FEGrid
    from ibvpy.mesh.fe_refinement_grid import FERefinementGrid
    from ibvpy.mesh.fe_domain import FEDomain
    from mathkit.mfn import MFnLineArray

    # Discretization
    fe_grid = FEGrid(coord_max=(10., 4., 0.),
                     shape=(10, 3),
                     fets_eval=fets_eval)

    bcg = BCDofGroup(var='u', value=0., dims=[0],
                     get_dof_method=fe_grid.get_left_dofs)
    bcg.setup(None)
    print('labels', bcg._get_labels())
    print('points', bcg._get_mvpoints())

    mf = MFnLineArray(  # xdata = arange(10),
        ydata=np.array([0, 1, 2, 3]))

    right_dof = 2
    tstepper = TS(sdomain=fe_grid,
                  bcond_list=[BCDofGroup(var='u', value=0., dims=[0, 1],
                                         get_dof_method=fe_grid.get_left_dofs),
                              #                                   BCDofGroup( var='u', value = 0., dims = [1],
                              # get_dof_method = fe_grid.get_bottom_dofs ),
                              BCDofGroup(var='u', value=.005, dims=[1],
                                         time_function=mf.get_value,
                                         get_dof_method=fe_grid.get_right_dofs)],
                  rtrace_list=[
                      RTDofGraph(name='Fi,right over u_right (iteration)',
                                 var_y='F_int', idx_y=right_dof,
                                 var_x='U_k', idx_x=right_dof,
                                 record_on='update'),
                      RTraceDomainListField(name='Stress',
                                            var='sig_app', idx=0,
                                            position='int_pnts',
                                            record_on='update'),
                      #                     RTraceDomainListField(name = 'Damage' ,
                      #                                    var = 'omega', idx = 0,
                      #
                      #                    record_on = 'update',
                      #                                    warp = True),
                      RTraceDomainListField(name='Displacement',
                                            var='u', idx=0,
                                            record_on='update',
                                            warp=True),
                      RTraceDomainListField(name='Strain energy',
                                            var='strain_energy', idx=0,
                                            record_on='update',
                                            warp=False),
                      RTraceDomainListInteg(name='Integ strain energy',
                                            var='strain_energy', idx=0,
                                            record_on='update',
                                            warp=False),
                      #                    RTraceDomainListField(name = 'N0' ,
                      #                                      var = 'N_mtx', idx = 0,
                      # record_on = 'update')
                  ]
                  )

    # Add the time-loop control
    #global tloop
    tloop = TLoop(tstepper=tstepper, KMAX=300, tolerance=1e-4,
                  tline=TLine(min=0.0, step=1.0, max=1.0))

    #import cProfile
    #cProfile.run('tloop.eval()', 'tloop_prof' )
    # print tloop.eval()
    #import pstats
    #p = pstats.Stats('tloop_prof')
    # p.strip_dirs()
    # print 'cumulative'
    # p.sort_stats('cumulative').print_stats(20)
    # print 'time'
    # p.sort_stats('time').print_stats(20)

    tloop.eval()
    # Put the whole thing into the simulation-framework to map the
    # individual pieces of definition into the user interface.
    #
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    app = IBVPyApp(ibv_resource=tloop)
    app.main()


if __name__ == '__main__':
    example_with_new_domain()



'''
Created on Feb 11, 2018

@author: rch
'''

import os

from mayavi.sources.vtk_xml_file_reader import VTKXMLFileReader
from tvtk.api import write_data

from ibvpy.tmodel.viz2d_field import Vis2DField, Viz2DField
import numpy as np
import traits.api as tr





from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic
from ibvpy.fets.fets2D.fets2D4q import FETS2D4Q
from ibvpy.fets.fets2D.fets2D4q8u import FETS2D4Q8U
from ibvpy.api import\
     BCDofGroup, TStepper as TS, TLoop, TLine, RTDofGraph
from ibvpy.rtrace.rt_domain_list_field import RTraceDomainListField
from ibvpy.mesh.fe_grid import FEGrid
from ibvpy.mesh.fe_refinement_grid import FERefinementGrid
from ibvpy.mesh.fe_domain import FEDomain

if __name__ == '__main__':

    fets_eval_4u = FETS2D4Q(mats_eval = MATS2DElastic())
    fets_eval_8u = FETS2D4Q8U(mats_eval = MATS2DElastic())
        
    fe_domain = FEDomain()

    fe_rgrid1 = FERefinementGrid( name = 'fe_rgrid1', fets_eval = fets_eval_4u, domain = fe_domain )

    fe_grid1 = FEGrid( name = 'fe_grid1', coord_max = (2.,6.,0.), 
                               shape   = (1,3),
                               fets_eval = fets_eval_4u,
                               level = fe_rgrid1 )    

    fe_grid2 = FEGrid( name = 'fe_grid2', coord_min = (2.,  6, 0.),
                      coord_max = (10, 15, 0.), 
                               shape   = (1,3),
                               fets_eval = fets_eval_4u,
                               level = fe_rgrid1 )    
        
    fe_rgrid2 = FERefinementGrid( name = 'fe_rgrid2', fets_eval = fets_eval_4u, domain = fe_domain )
    
    fe_grid3 = FEGrid( name = 'fe_grid3', coord_min = (0, 0, 1.),
                      coord_max = (2., 6.,1.), 
                               shape   = (1,3),
                               fets_eval = fets_eval_4u,
                               level = fe_rgrid2 )    

    fe_grid4 = FEGrid( name = 'fe_grid4', coord_min = (2.,  6, 1.),
                      coord_max = (10, 15, 1.), 
                               shape   = (1,3),
                               fets_eval = fets_eval_4u,
                               level = fe_rgrid2 )        

    fe_rgrid3 = FERefinementGrid( name = 'fe_rgrid3', fets_eval = fets_eval_4u, domain = fe_domain )
    
    fe_grid5 = FEGrid( name = 'fe_grid5', coord_min = (0, 0, 2.),
                      coord_max = (2., 6.,2.), 
                               shape   = (1,3),
                               fets_eval = fets_eval_4u,
                               level = fe_rgrid3 )                                   

    fe_grid6 = FEGrid( name = 'fe_grid6', coord_min = (2.,  6, 2.),
                      coord_max = (10, 15, 2.), 
                               shape   = (1,3),
                               fets_eval = fets_eval_4u,
                               level = fe_rgrid3 )                                   

    ts = TS( dof_resultants = True,
             sdomain = fe_domain,
             bcond_list =  [BCDofGroup(var='f', value = 1., dims = [0],
                                       get_dof_method = fe_grid1.get_top_dofs ),
                            BCDofGroup(var='u', value = 0., dims = [0,1],
                                       get_dof_method = fe_grid1.get_bottom_dofs ),
                                       ],
             rtrace_list =  [ RTDofGraph(name = 'Fi,right over u_right (iteration)' ,
                                   var_y = 'F_int', idx_y = 0,
                                   var_x = 'U_k', idx_x = 1),
                        RTraceDomainListField(name = 'Stress',
                             var = 'sig_app', idx = 0, warp = False ),
#                             RTraceDomainField(name = 'Displacement' ,
#                                        var = 'u', idx = 0),
#                                 RTraceDomainField(name = 'N0' ,
#                                              var = 'N_mtx', idx = 0,
#                                              record_on = 'update')
                    ]
                )
    
    
    # Add the time-loop control
    tloop = TLoop( tstepper = ts,
                   tline  = TLine( min = 0.0,  step = 1, max = 1.0 ))
    
    print(tloop.setup())
    
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    ibvpy_app = IBVPyApp( ibv_resource = tloop )
    ibvpy_app.main()
        


'''

@author: rch
'''

from threading import Thread
from ibvpy.view.reporter import Reporter
from ibvpy.view.reporter.reporter import ReportStudy
from .tmodel import TModel
from ibvpy.view.window.bmcs_viz_sheet import BMCSVizSheet
import traits.api as tr


class RunTimeLoopThread(Thread):
    '''Time loop thread responsible.
    '''

    def __init__(self, study, *args, **kw):
        super(RunTimeLoopThread, self).__init__(*args, **kw)
        self.daemon = True
        self.study = study

    def run(self):
        self.study.model.run()

    def pause(self):
        self.study.model.paused = True

    def stop(self):
        self.study.model.restart = True


class XBMCSStudy(ReportStudy):
    '''Combine the model with specification of outputs
    '''

    model = tr.Instance(TModel)
    '''Model of the studied phoenomenon.
    '''

    viz_sheet = tr.Instance(BMCSVizSheet, ())
    '''Sheet for 2d visualization.
    '''

    input = tr.Property

    def _get_input(self):
        return self.model

    output = tr.Property

    def _get_output(self):
        return self.viz_sheet

    offline = tr.DelegatesTo('viz_sheet')
    n_cols = tr.DelegatesTo('viz_sheet')

    def _model_changed(self):
        self.model.set_ui_recursively(self)
        tline = self.model.tline
        self.viz_sheet.time_range_changed(tline.max)
        self.viz_sheet.time_changed(tline.val)

    run_thread = tr.Instance(RunTimeLoopThread)

    running = tr.Bool(False)
    enable_run = tr.Bool(True)
    enable_pause = tr.Bool(False)
    enable_stop = tr.Bool(False)

    def _running_changed(self):
        '''If the simulation is running disable the run botton,
        enable the pause button and disable changes in all 
        input parameters.
        '''
        self.enable_run = not self.running
        self.enable_pause = self.running
        self.model.set_traits_with_metadata(self.enable_run,
                                            disable_on_run=True)

    start_event = tr.Event
    '''Event announcing the start of the calculation
    '''

    def _start_event_fired(self):
        print('START EVENT FIRED')
        self.viz_sheet.run_started()

    finish_event = tr.Event
    '''Event announcing the start of the calculation
    '''

    def _finish_event_fired(self):
        print('FINISH EVENT FIRED')
        self.viz_sheet.run_finished()

    def run(self):
        if self.running:
            return
        self.enable_stop = True
        self.run_thread = RunTimeLoopThread(self)
        self.run_thread.start()

    def join(self):
        '''Wait until the thread finishes
        '''
        self.run_thread.join()

    def pause(self):
        self.model.pause()

    def stop(self):
        self.model.stop()
        self.enable_stop = False

    def report_tex(self):
        r = Reporter(report_name=self.model.name,
                     input=self.model,
                     output=self.viz_sheet)
        r.write()
        r.show_tex()

    def report_pdf(self):
        r = Reporter(studies=[self])
        r.write()
        r.show_tex()
        r.run_pdflatex()
        r.show_pdf()

    def add_viz2d(self, clname, name, **kw):
        self.model.add_viz2d(clname, name, **kw)



'''
Created on Dec 3, 2015

@author: rch
'''

from os.path import join

from ibvpy.view.reporter import ROutputItem
from traits.api import \
    HasStrictTraits, Dict, Property, Float, \
    Bool, WeakRef, DelegatesTo, cached_property, \
    Str, List, Button

import matplotlib.pyplot as plt


class Viz2D(ROutputItem):
    '''Base class of the visualization adaptors
    '''

    viz_sheet = WeakRef

    name = Str('<unnamed>')
    visible = Bool(True)

    label = Property(depends_on='label')

    @cached_property
    def _get_label(self):
        return self.name
    vis2d = WeakRef

    def reset(self, ax):
        pass

    def clear(self):
        pass

    def plot(self, ax, vot=0):
        raise NotImplemented

    def plot_tex(self, ax, vot, *args, **kw):
        raise NotImplemented

    def write_figure(self, f, rdir, rel_study_path):
        fname = 'fig_' + self.name.replace(' ', '_') + '.pdf'
        f.write(r'''
\includegraphics[width=7.5cm]{%s}
''' % join(rel_study_path, fname))
        self.savefig(join(rdir, fname))

    def savefig_animate(self, vot, fname, figsize=(9, 6)):
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)
        self.plot(ax, vot)
        fig.tight_layout()
        fig.savefig(fname)

    def savefig(self, fname):
        fig = plt.figure()
        ax = fig.add_subplot(111)
        self.plot_tex(ax, 0.25)
        fig.savefig(fname)



from abc import ABC

from traits.api import HasTraits, Float

from numpy import where, zeros, dot, diag, linalg
from math import sqrt
from ibvpy.tmodel.mats3D.mats3D_tensor import map3d_sig_eng_to_mtx


class IStrainNorm3D(HasTraits):

    def get_f_trial(self, epsilon, D_el, E, nu, kappa):
        raise NotImplementedError

    def get_dede(self, epsilon, D_el, E, nu):
        raise NotImplementedError


class Euclidean(IStrainNorm3D):
    P_I = diag([1., 1., 1., 0.5, 0.5, 0.5])

    def get_f_trial(self, epsilon, D_el, E, nu, kappa):
        '''
        Returns equivalent strain - kappa
        @param epsilon:
        @param D_el:
        @param E:
        @param kappa:
        '''
        return sqrt(dot(dot(epsilon, self.P_I), epsilon.T)) - kappa


#    def get_dede(self, epsilon, D_el, E, nu):
#        dede = zeros(6)
#        t1 = pow(epsilon[0], 2.)
#        t3 = pow(epsilon[1], 2.)
#        t5 = pow(epsilon[2], 2.)
#        t7 = pow(epsilon[3], 2.)
#        t9 = pow(epsilon[4], 2.)
#        t11 = pow(epsilon[5], 2.)
#        t14 = sqrt(4. * t1 + 4. * t3 + 4. * t5 + 2. * t7 + 2. * t9 + 2. * t11)
#        t15 = 1. / t14
#        dede[0] = 2. * t15 * epsilon[0]
#        dede[1] = 2. * t15 * epsilon[1]
#        dede[2] = 2. * t15 * epsilon[2]
#        dede[3] = t15 * epsilon[3]
#        dede[4] = t15 * epsilon[4]
#        dede[5] = t15 * epsilon[5]
#        return dede


# from the maple shit damage_model_2D_euclidean_local_Code
class Energy(IStrainNorm3D, ABC):

    def get_f_trial(self, epsilon, D_el, E, nu, kappa):
        # print "time %8.2f sec"%diff
        return sqrt(1. / E * dot(dot(epsilon, D_el), epsilon.T)) - kappa


#    def get_dede(self, epsilon, D_el, E, nu):
#        dede = zeros(3)
#        t1 = 1. / E
#        t2 = t1 * epsilon[0]
#        t3 = t2 * D_el[0][0]
#        t4 = t1 * epsilon[1]
#        t5 = t4 * D_el[1][0]
#        t8 = t2 * D_el[0][1]
#        t9 = t4 * D_el[1][1]
#        t12 = pow(epsilon[2], 2.)
#        t16 = sqrt(epsilon[0] * (t3 + t5) + epsilon[1] * (t8 + t9) + t12 * t1 * D_el[2][2])
#        t17 = 1. / t16
#        dede[0] = t17 * (2. * t3 + t5 + t4 * D_el[0][1]) / 2.
#        dede[1] = t17 * (t2 * D_el[1][0] + t8 + 2. * t9) / 2.
#        dede[2] = t17 * t1 * epsilon[2] * D_el[2][2]
#        return dede


# from the maple shit damage_model_2D_von_Mises_local_Code
class Mises(IStrainNorm3D):
    k = Float(10.,
              label="k",
              desc="Shape Parameter")

    def get_f_trial(self, epsilon, D_el, E, nu, kappa):
        t1 = self.k - 1.
        t2 = (epsilon[0] + epsilon[1] + epsilon[2])
        t4 = 1.0 / self.k
        t6 = 1 - 2 * nu
        t10 = t1 ** 2
        t11 = t6 ** 2
        t14 = t2 ** 2
        t16 = epsilon[0] ** 2
        t18 = epsilon[1] ** 2
        t20 = epsilon[2] ** 2
        t22 = epsilon[3] ** 2
        t24 = epsilon[4] ** 2
        t26 = epsilon[5] ** 2
        t32 = (1 + nu) ** 2
        t37 = sqrt((t10 / t11 * t14) + 12. * self.k * (
                    t16 / 2. + t18 / 2. + t20 / 2. + t22 / 4. + t24 / 4. + t26 / 4. - t14 / 6.) / t32)
        t40 = (t1 * t2 * t4 / t6) / 2. + t4 * t37 / 2.
        return t40 - kappa

    def get_dede(self, epsilon, D_el, E, nu):
        dede = zeros(6)
        t1 = self.k - 1
        t2 = 1 / self.k
        t5 = 1 - 2 * nu
        t8 = (t1 * t2 / t5) / 2.
        t9 = t1 * t1
        t10 = t5 * t5
        t12 = t9 / t10
        t13 = epsilon[0] + epsilon[1] + epsilon[2]
        t14 = t13 * t13
        t16 = epsilon[0] ** 2
        t18 = epsilon[1] ** 2
        t20 = epsilon[2] ** 2
        t22 = epsilon[3] ** 2
        t24 = epsilon[4] ** 2
        t26 = epsilon[5] ** 2
        t32 = (1 + nu) ** 2
        t33 = 1 / t32
        t37 = sqrt((t12 * t14) + 0.12e2 * self.k * (
                    t16 / 2. + t18 / 2. + t20 / 2. + t22 / 4. + t24 / 4. + t26 / 4. - t14 / 6.) * t33)
        t38 = 1. / t37
        t39 = t2 * t38
        t41 = 2 * t12 * t13
        t43 = epsilon[1] / 3.
        t44 = epsilon[2] / 3.
        t54 = epsilon[0] / 3.
        dede[0] = t8 + t39 * (t41 + 12. * self.k * (2. / 3. * epsilon[0] - t43 - t44) * t33) / 4.
        dede[1] = t8 + t39 * (t41 + 12. * self.k * (2. / 3. * epsilon[1] - t54 - t44) * t33) / 4.
        dede[2] = t8 + t39 * (t41 + 12. * self.k * (2. / 3. * epsilon[2] - t54 - t43) * t33) / 4.
        dede[3] = 3. / 2. * t38 * epsilon[3] * t33
        dede[4] = 3. / 2. * t38 * epsilon[4] * t33
        dede[5] = 3. / 2. * t38 * epsilon[5] * t33
        return dede


class Rankine(IStrainNorm3D):
    '''
    computes main stresses and makes a norm of their positive part
    '''

    def get_f_trial(self, epsilon, D_el, E, nu, kappa):
        sigma_I = linalg.eigh(map3d_sig_eng_to_mtx(dot(D_el, epsilon)))[0]  # main stresses
        eps_eqv = linalg.norm(where(sigma_I >= 0., sigma_I, zeros(3))) / E  # positive part and norm
        return eps_eqv - kappa


#    def get_dede(self, epsilon, D_el, E, nu):
#        dede = zeros(3)
#        t1 = D_el[0][0] / 2.
#        t2 = D_el[1][0] / 2.
#        t3 = pow(D_el[0][0], 2.)
#        t4 = pow(epsilon[0], 2.)
#        t6 = D_el[0][0] * epsilon[0]
#        t7 = D_el[0][1] * epsilon[1]
#        t13 = D_el[1][1] * epsilon[1]
#        t16 = pow(D_el[0][1], 2.)
#        t17 = pow(epsilon[1], 2.)
#        t19 = D_el[1][0] * epsilon[0]
#        t25 = pow(D_el[1][0], 2.)
#        t29 = pow(D_el[1][1], 2.)
#        t31 = pow(D_el[2][2], 2.)
#        t32 = pow(epsilon[2], 2.)
#        t35 = t3 * t4 + 2 * t6 * t7 - 2. * D_el[0][0] * t4 * D_el[1][0] - 2. * t6 * t13 + t16 * t17 - 2. * t7 * t19 - 2. * D_el[0][1] * t17 * D_el[1][1] + t25 * t4 + 2. * t19 * t13 + t29 * t17 + 4. * t31 * t32
#        t36 = sqrt(t35)
#        t37 = 1. / t36
#        t57 = t37 * (2. * t3 * epsilon[0] + 2. * D_el[0][0] * D_el[0][1] * epsilon[1] - 4. * t6 * D_el[1][0] - 2. * D_el[0][0] * D_el[1][1] * epsilon[1] - 2. * t7 * D_el[1][0] + 2. * t25 * epsilon[0] + 2. * D_el[1][0] * D_el[1][1] * epsilon[1]) / 4.
#        t59 = fabs(t6 / 2. + t7 / 2. + t19 / 2. + t13 / 2. + t36 / 2.) / (t6 / 2. + t7 / 2. + t19 / 2. + t13 / 2. + t36 / 2.)
#        t63 = 1. / E
#        t66 = D_el[0][1] / 2.
#        t67 = D_el[1][1] / 2.
#        t85 = t37 * (2. * t6 * D_el[0][1] - 2. * t6 * D_el[1][1] + 2. * t16 * epsilon[1] - 2. * D_el[0][1] * D_el[1][0] * epsilon[0] - 0.4e1 * t7 * D_el[1][1] + 2. * t19 * D_el[1][1] + 2. * t29 * epsilon[1]) / 4.
#        dede[0] = (t1 + t2 + t57 + t59 * (t1 + t2 + t57)) * t63 / 2.
#        dede[1] = (t66 + t67 + t85 + t59 * (t66 + t67 + t85)) * t63 / 2.
#        dede[2] = (t37 * t31 * epsilon[2] + t59 * t37 * t31 * epsilon[2]) * t63
#        return dede


class Mazars(IStrainNorm3D):

    def get_f_trial(self, epsilon, D_el, E, nu, kappa):
        epsilon_pp = where(epsilon >= 0., epsilon, zeros(6))  # positive part
        return sqrt(1. / E * dot(dot(epsilon_pp, D_el), epsilon_pp.T)) - kappa



'''
Created on Feb 8, 2018

@author: rch
'''

from .fets2D import FETS2D
from ibvpy.mathkit.tensor import DELTA23_ab
import numpy as np
import sympy as sp
import traits.api as tr


#=================================================
# 4 nodes iso-parametric quadrilateral element
#=================================================
# generate shape functions with sympy
xi_1 = sp.symbols('xi_1')
xi_2 = sp.symbols('xi_2')


#=========================================================================
# Finite element specification
#=========================================================================

N_xi_i = sp.Matrix([(1.0 - xi_1) * (1.0 - xi_2) / 4.0,
                    (1.0 + xi_1) * (1.0 - xi_2) / 4.0,
                    (1.0 + xi_1) * (1.0 + xi_2) / 4.0,
                    (1.0 - xi_1) * (1.0 + xi_2) / 4.0], dtype=np.float64)

get_N_xi_i = sp.lambdify((xi_1, xi_2), N_xi_i, 'numpy')

dN_xi_ir = sp.Matrix(((-(1.0 / 4.0) * (1.0 - xi_2), -(1.0 / 4.0) * (1.0 - xi_1)),
                      ((1.0 / 4.0) * (1.0 - xi_2), -
                       (1.0 / 4.0) * (1.0 + xi_1)),
                      ((1.0 / 4.0) * (1.0 + xi_2), (1.0 / 4.0) * (1.0 + xi_1)),
                      (-(1.0 / 4.0) * (1.0 + xi_2), (1.0 / 4.0) * (1.0 - xi_1))), dtype=np.float64)

get_dN_xi_ir = sp.lambdify((xi_1, xi_2), dN_xi_ir, 'numpy')

class FETS2D4Q(FETS2D):
    dof_r = tr.Array(np.float64,
                     value=[[-1, -1], [1, -1], [1, 1], [-1, 1]])
    geo_r = tr.Array(np.float64,
                     value=[[-1, -1], [1, -1], [1, 1], [-1, 1]])
    vtk_r = tr.Array(np.float64,
                     value=[[-1, -1], [1, -1], [1, 1], [-1, 1]])
    n_nodal_dofs = 2
    vtk_r = tr.Array(np.float64, value=[[-1, -1], [1, -1], [1, 1],
                                       [-1, 1]])
    vtk_cells = [[0, 1, 2, 3]]
    vtk_cell_types = 'Quad'
    vtk_cell = [0, 1, 2, 3]
    vtk_cell_type = 'Quad'

    vtk_expand_operator = tr.Array(np.float64, value=DELTA23_ab)

    # numerical integration points (IP) and weights
    xi_m = tr.Array(np.float64,
                    value=[[-1.0 / np.sqrt(3.0), -1.0 / np.sqrt(3.0)],
                           [1.0 / np.sqrt(3.0), -1.0 / np.sqrt(3.0)],
                           [1.0 / np.sqrt(3.0), 1.0 / np.sqrt(3.0)],
                           [-1.0 / np.sqrt(3.0), 1.0 / np.sqrt(3.0)]
                           ])

    w_m = tr.Array(value=[1, 1, 1, 1], dtype=np.float64)

    n_m = tr.Property

    def _get_n_m(self):
        return len(self.w_m)

    shape_function_values = tr.Property(tr.Tuple)
    '''The values of the shape functions and their derivatives at the IPs
    '''
    @tr.cached_property
    def _get_shape_function_values(self):
        N_mi = np.array([N_xi_i.subs(list(zip([xi_1, xi_2], xi)))
                         for xi in self.xi_m], dtype=np.float64)[...,0]
        N_im = np.einsum('mi->im', N_mi)
        dN_mir = np.array([dN_xi_ir.subs(list(zip([xi_1, xi_2], xi)))
                           for xi in self.xi_m], dtype=np.float64).reshape(4, 4, 2)
        dN_nir = np.array([dN_xi_ir.subs(list(zip([xi_1, xi_2], xi)))
                           for xi in self.vtk_r], dtype=np.float64).reshape(4, 4, 2)
        dN_imr = np.einsum('mir->imr', dN_mir)
        dN_inr = np.einsum('nir->inr', dN_nir)
        return (N_im, dN_imr, dN_inr)

    # shape_function_values = tr.Property(tr.Tuple)
    # '''The values of the shape functions and their derivatives at the IPs
    # '''
    # @tr.cached_property
    # def _get_shape_function_values(self):
    #     N_mi = get_N_xi_i(*self.xi_m.T)
    #     N_im = np.einsum('mi->im', N_mi)
    #     dN_mir = get_dN_xi_ir(*self.xi_m.T)
    #     dN_nir = get_dN_xi_ir(*self.xi_n.T)
    #     dN_imr = np.einsum('mir->imr', dN_mir)
    #     dN_inr = np.einsum('nir->inr', dN_nir)
    #     return (N_im, dN_imr, dN_inr)

    N_im = tr.Property()
    '''Shape function values in integration poindots.
    '''

    def _get_N_im(self):
        return self.shape_function_values[0]

    dN_imr = tr.Property()
    '''Shape function derivatives in integration poindots.
    '''

    def _get_dN_imr(self):
        return self.shape_function_values[1]

    dN_inr = tr.Property()
    '''Shape function derivatives in visualization poindots.
    '''

    def _get_dN_inr(self):
        return self.shape_function_values[2]


if __name__ == '__main__':
    fe = FETS2D4Q()
    fe.configure_traits()



from numpy import \
     zeros, dot, hstack, vstack, identity, cross, transpose, tensordot, outer

from scipy.linalg import \
     inv, norm

from ibvpy.tmodel.mats2D.mats2D_tensor import \
    map2d_eps_mtx_to_eng, map2d_sig_eng_to_mtx

from ibvpy.tmodel.mats3D.mats3D_tensor import \
    map3d_eps_eng_to_mtx, map3d_sig_eng_to_mtx, map3d_sig_mtx_to_eng, map3d_eps_mtx_to_eng, map3d_tns2_to_tns4, map3d_tns4_to_tns2, map3d_eps_mtx_to_eng

def get_Trans_strain_from_glb_to_loc(dircos_mtx):
    '''Transformation matrix in 3d shape = (6,6).
    (eps_mtx_loc = T * eps_mtx_glb * T^t)
    Multiply the global strain vector 'eps_vct_glb' (in engineering notation) with the returned
    transformation matrix 'Trans_strain_from_glb_to_loc_mtx' in order to obtain the loacal strain vector
    (in engineering notation) in the local coordinate system (x',y',z'), e.g. in-plan shell directions.
    The ordering of the strain vector in engineering notation corresponds to the VOIGT-notation:
    eps_glb = [eps_xx, eps_yy, eps_zz, eps_yz, eps_zx, eps_xz].
    The parameter 'dircos_mtx' is the matrix of the direction cosines describing the relation between
    the global and local coordinate system.
    (the returned transformation matrix corresponds to Zienkiewicz Eq.(6.24) but used VOIGT-notation instead)
    '''
    t = dircos_mtx  # shape=(3,3)
    T = zeros((6, 6))  # shape=(6,6)
    T[0, 0] = t[0, 0] * t[0, 0]
    T[0, 1] = t[0, 1] * t[0, 1]
    T[0, 2] = t[0, 2] * t[0, 2]
    T[0, 3] = t[0, 1] * t[0, 2]
    T[0, 4] = t[0, 0] * t[0, 2]
    T[0, 5] = t[0, 0] * t[0, 1]
    T[1, 0] = t[1, 0] * t[1, 0]
    T[1, 1] = t[1, 1] * t[1, 1]
    T[1, 2] = t[1, 2] * t[1, 2]
    T[1, 3] = t[1, 1] * t[1, 2]
    T[1, 4] = t[1, 0] * t[1, 2]
    T[1, 5] = t[1, 0] * t[1, 1]
    T[2, 0] = t[2, 0] * t[2, 0]
    T[2, 1] = t[2, 1] * t[2, 1]
    T[2, 2] = t[2, 2] * t[2, 2]
    T[2, 3] = t[2, 1] * t[2, 2]
    T[2, 4] = t[2, 0] * t[2, 2]
    T[2, 5] = t[2, 0] * t[2, 1]
    T[3, 0] = 2.0 * t[2, 0] * t[1, 0]
    T[3, 1] = 2.0 * t[2, 1] * t[1, 1]
    T[3, 2] = 2.0 * t[2, 2] * t[1, 2]
    T[3, 3] = t[2, 1] * t[1, 2] + t[2, 2] * t[1, 1]
    T[3, 4] = t[2, 2] * t[1, 0] + t[2, 0] * t[1, 2]
    T[3, 5] = t[2, 0] * t[1, 1] + t[2, 1] * t[1, 0]
    T[4, 0] = 2.0 * t[2, 0] * t[0, 0]
    T[4, 1] = 2.0 * t[2, 1] * t[0, 1]
    T[4, 2] = 2.0 * t[2, 2] * t[0, 2]
    T[4, 3] = t[2, 1] * t[0, 2] + t[2, 2] * t[0, 1]
    T[4, 4] = t[2, 2] * t[0, 0] + t[2, 0] * t[0, 2]
    T[4, 5] = t[2, 0] * t[0, 1] + t[2, 1] * t[0, 0]
    T[5, 0] = 2.0 * t[1, 0] * t[0, 0]
    T[5, 1] = 2.0 * t[1, 1] * t[0, 1]
    T[5, 2] = 2.0 * t[1, 2] * t[0, 2]
    T[5, 3] = t[1, 1] * t[0, 2] + t[1, 2] * t[0, 1]
    T[5, 4] = t[1, 2] * t[0, 0] + t[1, 0] * t[0, 2]
    T[5, 5] = t[1, 0] * t[0, 1] + t[1, 1] * t[0, 0]
    Trans_strain_to_loc_mtx = T
    return Trans_strain_to_loc_mtx

def get_Trans_stress_from_loc_to_glb(dircos_mtx):
    '''Transformation matrix in 3d shape = (6,6)
    (sig_mtx_glb = T^t * sig_mtx_loc * T)
    Multiply the local stress vector 'sig_vct_loc' (in engineering notation in the local coordinate
    system (x',y', z') with the returned transformation matrix 'Trans_stress_from_loc_to_glb_mtx' in
    order to obtain the global stress vector 'sig_vct_glb' (in engineering notation) in the global
    coordinate system (x,y,z). (cf. Zienkiewicz Eq.(6.23) multiplied with T_t*(...)*T, for T being an orthonormal basis)
    The ordering of the stress vector in engineering notation corresponds to the VOIGT-notation:
    sig_vct_glb = [sig_xx, sig_yy, sig_zz, sig_yz, sig_zx, sig_xz].
    The parameter 'dircos_mtx' is the matrix of the direction cosines describing the relation between
    the global and the local coordinate system.
    '''
    t = dircos_mtx  # shape=(3,3)
    T = zeros((6, 6))  # shape=(6,6)
    T[0, 0] = t[0, 0] * t[0, 0]
    T[0, 1] = t[1, 0] * t[1, 0]
    T[0, 2] = t[2, 0] * t[2, 0]
    T[0, 3] = 2.0 * t[1, 0] * t[2, 0]
    T[0, 4] = 2.0 * t[0, 0] * t[2, 0]
    T[0, 5] = 2.0 * t[0, 0] * t[1, 0]
    T[1, 0] = t[0, 1] * t[0, 1]
    T[1, 1] = t[1, 1] * t[1, 1]
    T[1, 2] = t[2, 1] * t[2, 1]
    T[1, 3] = 2.0 * t[1, 1] * t[2, 1]
    T[1, 4] = 2.0 * t[0, 1] * t[2, 1]
    T[1, 5] = 2.0 * t[0, 1] * t[1, 1]
    T[2, 0] = t[0, 2] * t[0, 2]
    T[2, 1] = t[1, 2] * t[1, 2]
    T[2, 2] = t[2, 2] * t[2, 2]
    T[2, 3] = 2.0 * t[1, 2] * t[2, 2]
    T[2, 4] = 2.0 * t[0, 2] * t[2, 2]
    T[2, 5] = 2.0 * t[0, 2] * t[1, 2]
    T[3, 0] = t[0, 2] * t[0, 1]
    T[3, 1] = t[1, 2] * t[1, 1]
    T[3, 2] = t[2, 2] * t[2, 1]
    T[3, 3] = t[1, 2] * t[2, 1] + t[2, 2] * t[1, 1]
    T[3, 4] = t[2, 2] * t[0, 1] + t[0, 2] * t[2, 1]
    T[3, 5] = t[0, 2] * t[1, 1] + t[1, 2] * t[0, 1]
    T[4, 0] = t[0, 2] * t[0, 0]
    T[4, 1] = t[1, 2] * t[1, 0]
    T[4, 2] = t[2, 2] * t[2, 0]
    T[4, 3] = t[1, 2] * t[2, 0] + t[2, 2] * t[1, 0]
    T[4, 4] = t[2, 2] * t[0, 0] + t[0, 2] * t[2, 0]
    T[4, 5] = t[0, 2] * t[1, 0] + t[1, 2] * t[0, 0]
    T[5, 0] = t[0, 1] * t[0, 0]
    T[5, 1] = t[1, 1] * t[1, 0]
    T[5, 2] = t[2, 1] * t[2, 0]
    T[5, 3] = t[1, 1] * t[2, 0] + t[2, 1] * t[1, 0]
    T[5, 4] = t[2, 1] * t[0, 0] + t[0, 1] * t[2, 0]
    T[5, 5] = t[0, 1] * t[1, 0] + t[1, 1] * t[0, 0]
    Trans_stress_from_loc_mtx = T
    return Trans_stress_from_loc_mtx




import copy

from traits.api import \
    HasStrictTraits, Bool, \
    Property, Instance, cached_property, \
    Dict, DelegatesTo, WeakRef, Array, Str

import numpy as np

from .i_tmodel import ITModel
from .i_xmodel import IXModel


class DomainState(HasStrictTraits):
    '''
    Container of spatial domains represented according to XModel.
    and corresponding state derived usi ng the TModel instance
    '''

    tstep = WeakRef

    hist = DelegatesTo('tstep')

    xmodel = Instance(IXModel)

    tmodel = Instance(ITModel)

    state_n = Property(Dict(Str, Array),
                       depends_on='model_structure_changed')
    '''Dictionary of state arrays.
    The entry names and shapes are defined by the material
    model.
    '''
    @cached_property
    def _get_state_n(self):
        xmodel_shape = self.xmodel.state_var_shape
        tmodel_shapes = self.tmodel.state_var_shapes
        return {
            name: np.zeros(xmodel_shape + mats_sa_shape, dtype=np.float64)
            for name, mats_sa_shape
            in list(tmodel_shapes.items())
        }

    hidden = Bool(False)

    state_k = Dict
    '''State variables within the current iteration step
    '''

    def get_corr_pred(self, U_k, t_n, t_n1):
        U_k_field = self.xmodel.map_U_to_field(U_k)
        self.state_k = copy.deepcopy(self.state_n) # todo: optimize memory management.
        sig_k, D_k = self.tmodel.get_corr_pred(
            U_k_field, t_n1, **self.state_k
        )
        K_k = self.xmodel.map_field_to_K(D_k)
        dof_E, f_Ei = self.xmodel.map_field_to_F(sig_k)
        return f_Ei, K_k, dof_E

    def record_state(self):
        '''The trial state k becomes a fundamental state n.
        '''
        for name, s_k in self.state_k.items():
            self.state_n[name] = s_k
        return self.state_n





# 2D5 elements - 3D elements with distinguished midface
from .fets2D58h import FETS2D58H
from .fets2D58h20u import FETS2D58H20U
from .fets2D58h16u import FETS2D58H16U
from .fets2D58h24u import FETS2D58H24U
from .fets2D58h32u import FETS2D58H32U





from ibvpy.tmodel.mats_eval import \
    IMATSEval, MATSEval
from traits.api import  \
    provides, Property, cached_property
from bmcs_utils.api import \
    Float, View, Item

import numpy as np

@provides(IMATSEval)
class MATS3DIfcElastic(MATSEval):

    node_name = "Elastic interface"

    E_n = Float(1e+6, tooltip='Normal stiffness of the interface [MPa]',
                MAT=True, unit='MPa/m', symbol='E_\mathrm{n}',
                desc='Normal stiffness of the interface',
                auto_set=False, enter_set=True)

    E_s = Float(1.0, tooltip='Shear stiffness of the interface [MPa]',
                MAT=True, unit='MPa/m', symbol='E_\mathrm{s}',
                desc='Shear-modulus of the interface',
                auto_set=True, enter_set=True)

    state_var_shapes = {}

    D_rs = Property(depends_on='E_n,E_s')

    @cached_property
    def _get_D_rs(self):
        return np.array([[self.E_n, 0, 0],
                         [0, self.E_s, 0],
                         [0, 0, self.E_s]], dtype=np.float64)

    def get_corr_pred(self, u_r, tn1):
        tau = np.einsum(
            'rs,...s->...r',
            self.D_rs, u_r)
        grid_shape = tuple([1 for _ in range(len(u_r.shape[:-1]))])
        D = self.D_rs.reshape(grid_shape + self.D_rs.shape)
        return tau, D

    ipw_view = View(
        Item('E_s'),
        Item('E_n')
    )



'''
Created on Feb 11, 2018

@author: rch
'''

import tempfile
import traits.api as tr


class Vis3D(tr.HasTraits):

    sim = tr.WeakRef

    tstep = tr.WeakRef

    dir = tr.Directory

    file_list = tr.List(tr.Str,
                        desc='a list of files belonging to a time series')

    extension = tr.Str('vtu')
    # vtk format file extension

    def setup(self):
        pass

    def new_dir(self):
        self.dir = tempfile.mkdtemp()

    def add_file(self, fname):
        self.file_list.append(fname)


class Viz3D(tr.HasTraits):

    vis3d = tr.WeakRef

    name = tr.Property()

    def _get_name(self):
        return self.vis3d.var

    def setup(self):
        pass

    def plot(self, vot):
        pass



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 18, 2009 by: rch

from .mats2D_cmdm import \
    MATS2DMicroplaneDamage, PhiFnStrainSoftening

from ibvpy.tmodel.mats2D.__test__ import TestMATS2D

class TestMATS2DMicroplaneDamageCompliance( TestMATS2D ):
    '''Instantiation of the material tests in 2D for the compliance version.
    '''
    def _mats_eval_default(self):
        return MATS2DMicroplaneDamage( E = 34000, nu = 0.25, 
                                       stress_state  = "plane_strain",
                                       model_version = 'compliance',
                                       phi_fn = PhiFnStrainSoftening(
                                                                  G_f = 0.001117,
                                                                  f_t = 2.8968,
                                                                  md = 0.0,
                                                                  h = 1.0 )                                       
                                       )

    # Tests for the stiffness version
    #
    def test_symmetry_tension(self):
        '''Symmetry of response for tensile and shear loading.
        '''
        self.assert_symmetry( load_dirs = [0,1] )
    
    def test_symmetry_shear(self):
        '''Symmetry of response for tensile and shear loading.
        '''
        self.assert_symmetry( load_dirs = [1,0] )

    def test_stress_value(self):
        '''Assert a correct value of resulting stress a material point.
        '''
        self.assert_stress_value( [2.43263348, 0, 0, 0],
                                    n_steps = 3, load = 0.0001 )

class TestMATS2DMicroplaneDamageStiffness( TestMATS2D ):
    '''Instantiation of the material tests in 2D for the stiffness version.
    '''
    def _mats_eval_default(self):
        return MATS2DMicroplaneDamage( E = 34000, nu = 0.25, 
                                       stress_state  = "plane_strain",
                                       model_version = 'stiffness',
                                       phi_fn = PhiFnStrainSoftening(
                                                                  G_f = 0.001117,
                                                                  f_t = 2.8968,
                                                                  md = 0.0,
                                                                  h = 1.0 )
                                       )

    # Tests for the stiffness version
    #
    def test_symmetry_tension_stiffness(self):
        '''Symmetry of response for tensile and shear loading.
        '''
        self.assert_symmetry( load_dirs = [0,1] )
    
    def test_symmetry_shear_stiffness(self):
        '''Symmetry of response for tensile and shear loading.
        '''
        self.assert_symmetry( load_dirs = [1,0] )

    def test_stress_value(self):
        '''Assert a correct value of resulting stress a material point.
        '''
        self.assert_stress_value( [2.541103649207904, 0, 0, 0],
                                  n_steps = 3, load = 0.0001 )

    def test_total_strain_energy_value(self):
        '''Assert a correct value of resulting elastic energy in the unit domain
        '''
        self.assert_total_energy_value( 2.541103649207904/2.*0.0001,
                                               n_steps = 3, load = 0.0001 )

    def test_total_fracture_energy_value(self):
        '''Assert a correct value of resulting elastic energy in the unit domain
        '''
        self.assert_total_energy_value( 0.00081182249919380263,
                                        ivar = 'fracture_energy',
                                        n_steps = 3, load = 0.01 )

if __name__ == "__main__":
    import unittest
    import sys;sys.argv = ['', 'TestMATS2DMicroplaneDamageStiffness.test_symmetry_shear_stiffness']    
    unittest.main()



from numpy import \
    array, zeros, array_equal, arange, \
    hstack, sort, where, unique, copy, frompyfunc, sign, fabs as nfabs, \
    logical_and, argwhere, fabs
from scipy.optimize import \
    brentq
from traits.api import \
    Instance, Property, cached_property, \
    provides, Float, \
    Callable, Str, Enum

from ibvpy.fets import FETSEval
from ibvpy.mesh.i_fe_grid_slice import IFEGridSlice
from ibvpy.mesh.i_fe_uniform_domain import IFEUniformDomain

from .fe_grid import MElem
from .fe_subdomain import FESubDomain


#--------------------------------------------------------------------------
# Summation procedure on an n-dimensional grid
#--------------------------------------------------------------------------
def edge_sum(b_arr):
    '''Perform summation of values in neighboring nodes
    along edges in each dimension.

    The original array [n,n,n] is reduced to [n-1,n-1,n-1] 
    '''
    n_dims = len(b_arr.shape)

    # construct the slice template - all indices are running through
    s_list = [
        [slice(None)
         for i in range(n_dims)
         ]
        for j in range(n_dims)
    ]

    # construct the left and right index
    # sl = :-1   - from first to second last
    # sr = 1:    - from second to last
    sl, sr = slice(None, -1), slice(1, None)

    # construct a slice of the type
    # b_arr[ :, :-1, : ] + b_arr[ :, 1:, : ]
    # performing the neighbor summation in all dimensions
    #
    for i, s in enumerate(s_list):
        slice_l, slice_r = copy(s), copy(s)
        slice_l[i] = sl
        slice_r[i] = sr
        b_arr = b_arr[tuple(slice_l)] + b_arr[tuple(slice_r)]

    return b_arr


def get_intersect_pt(fn, args):
    try:
        return brentq(fn, -1, 1, args=args)
    except ValueError:
        return


@provides(IFEUniformDomain)
class XFESubDomain(FESubDomain):
    '''Subgrid derived from another grid domain.
    '''

    # Distinguish which part of the level set to take
    #
    domain_tag = Enum('both', 'pos', 'neg')
    rt_tol = Float(0.0001)

    # specialized label
    _tree_label = Str('extension domain')

    # container domain (why is this needed?)
    _domain = Instance('ibvpy.mesh.fe_domain.FEDomain')
    domain = Property

    def _set_domain(self, value):
        'reset the domain of this domain'
        if self._domain:
            # unregister in the old domain
            raise NotImplementedError(
                'FESubDomain cannot be relinked to another FEDomain')

        self._domain = value
        # register in the domain as a subdomain
        self._domain.xdomains.append(self)
        self._domain._append_in_series(self)

    def _get_domain(self):
        return self._domain

    #-----------------------------------------------------------------
    # Associated time stepper
    #-----------------------------------------------------------------

    # element level
    fets_eval = Instance(FETSEval)

    # domain level
    dots = Property

    @cached_property
    def _get_dots(self):
        '''Construct and return a new instance of domain
        time stepper.
        '''
        return XDOTSEval(sdomain=self)

    #-----------------------------------------------------------------
    # Boundary level set - define the boundary of the subdomain
    #-----------------------------------------------------------------
    # Specify the boundary of the subdomain
    # The elements outside of the domain are
    # not included in the integration and are
    # not deactivated from the original mesh.
    #
    bls_function = Callable

    def _bls_function_default(self):
        def unity(*x):
            # by default the level set function is positive in the whole domain
            # - the level set is valid in the whole domain
            return sign(fabs(x[0]))
        return unity

    #-----------------------------------------------------------------
    # Level set separating the domain into the positive and negative parts
    #-----------------------------------------------------------------

    # current slice instance
    _slice = Instance(IFEGridSlice)

    # slice property
    slice = Property

    def _set_slice(self, new_slice):
        '''Set the slice by keeping track of the elements included in the subdomain so far. 
        '''
        # remember the element numbers that were in the slice so far.

        old_state_array = None

        if self._slice:
            old_elems = self.elems
            # get the new elemes
            new_elems = new_slice.elems
            # remember the state associated with the elements
            print('old_elems', old_elems)
            print('new_elems', new_elems)
            old_size = old_elems.size
            new_size = new_elems.size

            # if the old array is contained at the beginning of the new array
            # reuse the state array
            if old_size < new_size and array_equal(old_elems, new_elems[:old_size]):
                old_state_array = copy(self.dots.state_array)

        self._slice = new_slice

        # reuse the old state array
        if old_state_array:
            self.dots.state_array[:old_size] = old_state_array

    def _get_slice(self):
        '''Get current slice.
        '''
        return self._slice

    fe_grid = Property

    def _get_fe_grid(self):
        return self._slice.fe_grid

    #---------------------------------------------------------------------
    # Element groups accessed as properties
    #---------------------------------------------------------------------

    # Elements cut by the boundary LS function and by the discontinuity LS
    #
    boundary_elems = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_boundary_elems(self):
        #
        # boundary elements are defined as cells that are
        # contain both positive and negative values of the level set function
        # and are crossed with the domain boundary.
        #
        b_elems_grid = logical_and(nfabs(self.bls_elem_grid) < 4,
                                   nfabs(self.ls_elem_grid) < 4)

        return self._get_elems_from_elem_grid(b_elems_grid)

    interior_elems = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_interior_elems(self):

        # evaluate the boundary operator - with a result boolean array
        # in the corner nodes.
        #
        b_elem_grid = self.bls_elem_grid

        ls_elem_grid = self.ls_elem_grid

        # make the intersection
        #
        elem_grid = logical_and(nfabs(ls_elem_grid) < 4, b_elem_grid == 4)

        return self._get_elems_from_elem_grid(elem_grid)

    def _get_elems_from_elem_grid(self, elem_grid):
        '''Get an array of element offsets based on the from an array
        with boolean identifiers of elements. All elements corresponding
        to the boolean identifiers are returned. Subsequently, the
        array based properties of the domain, like elem_dof_map,
        elem_geo_x and elem_X_geo can be accessed.
        '''
        # get the mask of elements belonging to the level set domain
        #
        elem_idx = argwhere(elem_grid)

        n_dims = elem_idx.shape[1]

        # prepare the index as a sequence of arrays in x, y, and z dimensions
        #
        elem_idx_tuple = tuple([elem_idx[:, i] for i in range(n_dims)])

        # get the enumeration of cells in a n-dim array
        #
        c_g = self.fe_grid.dof_grid.cell_grid.cell_idx_grid

        # select the elements cut by the level set
        #
        return c_g[elem_idx_tuple]

    elems = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_elems(self):
        return hstack([self.interior_elems, self.boundary_elems])

    bls_elem_grid = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_bls_elem_grid(self):

        X_grid = self.fe_grid.dof_grid.cell_grid.vertex_X_grid

        # evaluate the boundary operator - with a result boolean array
        # in the corner nodes.
        #
        b_vertex_grid = sign(self.bls_function(*X_grid))

        return edge_sum(b_vertex_grid)

    ls_elem_grid = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_ls_elem_grid(self):

        X_grid = self.fe_grid.dof_grid.cell_grid.vertex_X_grid

        # find the indices of the elems

        ls_function = self.slice.ls_function

        n_dims = X_grid.shape[0]

        vect_fn = frompyfunc(ls_function, n_dims, 1)
        ls_vertex_grid = sign(vect_fn(*X_grid))

        return array(edge_sum(ls_vertex_grid), dtype=float)

    def deactivate_sliced_elems(self):
        for e in self.elems:
            self.fe_grid.deactivate(e)

    elements = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_elements(self):
        elem_dof_map = self.elem_dof_map
        elem_X_map = self.elem_X_map
        elem_x_map = self.elem_x_map
        return [MElem(dofs=dofs, point_X_arr=point_X_arr, point_x_arr=point_x_arr)
                for dofs, point_X_arr, point_x_arr
                in zip(elem_dof_map, elem_X_map, elem_x_map)]

    elem_dof_map = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_elem_dof_map(self):
        return self.dof_enum_data[0]

    dof_enum_data = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_dof_enum_data(self):
        '''
        generates  elem_dof_map and n_dofs
        TODO - optimize the the algorithm, replace loop with sucessive slices
        to find jump in the enumeration
        TODO - generalize the algorithm from dofs to nodes, this will allow
        to have more extended dofs than parent ones  
        '''
        # (elems, dofs)

        # @todo
        # It would be better to provide fe_grid with
        # several element access methods - specify an array of elements
        # and you get a consistent array of dofs, geo_x and geo_X values.
        #
        dofs = self.fe_grid.dof_grid.cell_dof_map[self.elems]

        sh = dofs.shape

        # number of added dofs
        n_xdofs = self.fets_eval.n_e_dofs - self.fets_eval.parent_fets.n_e_dofs

        # This is ugly hack for rough version
        # @todo
        # the number of enriching dofs must be specified explicitly in
        # the fets_eval - used for enrichment.
        #
        if (n_xdofs / 2) > self.fets_eval.parent_fets.n_e_dofs:
            #(array([[0,1,2,3,4,5,6,7,8,9,10,11]]),10)
            return (array([[1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]]), 10)

        # x_slice stands for number of enriched dofs in a single node.
        #
        elem_dof_map = copy(dofs).reshape(
            sh[0], self.fets_eval.parent_fets.n_e_dofs)

        elem_xdof_map = (copy(dofs)[:, :, self.fets_eval.x_slice]).reshape(
            sh[0], n_xdofs)

        # special case for selective integration - no enrichment. returns
        # the original dofs and zero n_dofs
        #
        if elem_xdof_map.shape[1] == 0:
            return (dofs, 0)  # without additional dofs

        # compress the dof enumeration to be a consecutive sequence
        # of integers without gaps
        #
        xidx = sort(unique(elem_xdof_map.flatten()))
        last_idx = 0
        for i in range(len(xidx)):
            shrink = xidx[i] - last_idx
            if shrink > 0:
                # shrink everything behind shrink
                elem_xdof_map[where(elem_xdof_map > i)] -= shrink
            last_idx = xidx[i] + 1

        elem_xdof_map += self.dof_offset
        #
        # return an array of dof numbers - the original dofs of the parent
        # domain are reused by the xfe_subdomain.
        #
        return (hstack([elem_dof_map, elem_xdof_map]), xidx.shape[0])

    elem_X_map = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_elem_X_map(self):
        geo_X = self.fe_grid.geo_grid.elem_X_map[self.elems]
        return geo_X

    elem_x_map = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_elem_x_map(self):
        geo_x = self.fe_grid.geo_grid.elem_x_map[self.elems]
        return geo_x

    dof_node_ls_values = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_dof_node_ls_values(self):

        X_grid = self.fe_grid.dof_grid.cell_grid.point_X_arr
        ls_function = self.slice.ls_function
        n_dims = X_grid.shape[1]

        vect_fn = frompyfunc(ls_function, n_dims, 1)
        ls_vertex_grid = sign(vect_fn(*X_grid.T))

        # select the affected elements from the ls_vertex_grid

        elem_node_map = self.fe_grid.dof_grid.cell_grid.cell_node_map

        return ls_vertex_grid[elem_node_map[self.elems]]

    ls_intersection_r = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_ls_intersection_r(self):
        '''Get the intersection points in local coordinates
        '''
        # fe_grid_slice.r_i # intersecting points
        fe_grid = self.fe_grid
        i_elements = self.elems
        el_pnts = []
        for elem in i_elements:
            inter_pts = []
            # X_mtx = self.elements[elem].get_X_mtx() # skips deactivated
            # elements
            X_mtx = fe_grid.geo_grid.get_cell_point_X_arr(elem)
            dim = X_mtx.shape[1]  # TODO:merge 1 and 2d
            if dim == 1:
                r_coord = get_intersect_pt(self.ls_fn_r, (0., X_mtx))
                if r_coord != None:
                    inter_pts.append([r_coord])
            elif dim == 2:
                for c_coord in [-1., 1.]:
                    args = (c_coord, X_mtx)
                    s_coord = get_intersect_pt(self.ls_fn_s, args)
                    r_coord = get_intersect_pt(self.ls_fn_r, args)
                    if s_coord != None:
                        inter_pts.append([c_coord, s_coord])
                    if r_coord != None:
                        inter_pts.append([r_coord, c_coord])
            elif dim == 3:
                raise NotImplementedError('not available for 3D yet')
            el_pnts.append(inter_pts)
        return array(el_pnts)

    def ls_fn_r(self, r, s, X_mtx):  # TODO:dimensionless treatment
        ls_function = self.slice.ls_function
        X_pnt = self.fe_grid.fets_eval.map_r2X([r, s], X_mtx)
        if X_pnt.shape[0] == 1:
            Y = 0.
        else:
            Y = X_pnt[1]
        return ls_function(X_pnt[0], Y)

    def ls_fn_s(self, s, r, X_mtx):
        ls_function = self.slice.ls_function
        X_pnt = self.fe_grid.fets_eval.map_r2X([r, s], X_mtx)
        return ls_function(X_pnt[0], X_pnt[1])

    n_elems = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_n_elems(self):
        return self.elems.shape[0]

    n_dofs = Property

    @cached_property
    def _get_n_dofs(self):
        #n_xdofs = self.fets_eval.n_e_dofs - self.fets_eval.parent_fets.n_e_dofs
        # return n_xdofs *  self.n_active_elems
        return self.dof_enum_data[1]

    idx_active_elems = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_idx_active_elems(self):
        return arange(self.shape)

    n_active_elems = Property(depends_on='_slice, boundary')

    @cached_property
    def _get_n_active_elems(self):
        return self.shape  # TODO:

    def apply_on_ip_grid(self, fn, ip_mask):
        '''
        Apply the function fn over the first dimension of the array.
        @param fn: function to apply for each ip from ip_mask and each element. 
        @param ip_mask: specifies the local coordinates within the element.     
        '''
        X_el = self.elem_X_map
        # test call to the function with single output - to get the shape of
        # the result.
        out_single = fn(ip_mask[0], X_el[0])
        out_grid_shape = (X_el.shape[0], ip_mask.shape[0], ) + out_single.shape
        out_grid = zeros(out_grid_shape)

        for el in range(X_el.shape[0]):
            for ip in range(ip_mask.shape[0]):
                out_grid[el, ip, ...] = fn(ip_mask[ip], X_el[el])

        return out_grid



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 18, 2009 by: rch

from .mats2D_elastic import \
    MATS2DElastic

from ibvpy.tmodel.mats2D.__test__ import TestMATS2D

class TestMATS2DElastic( TestMATS2D ):

    def _mats_eval_default(self):
        return MATS2DElastic( E = 34000, nu = 0.25 ) 

    def test_symmetry_tension(self):
        '''Symmetry of response for tensile and shear loading.
        '''
        self.assert_symmetry( load_dirs = [0,1] )
    
    def test_symmetry_shear(self):
        '''Symmetry of response for tensile and shear loading.
        '''
        self.assert_symmetry( load_dirs = [1,0] )

    def test_stress_value(self):
        '''Assert a correct value of resulting stress a material point.
        '''
        self.assert_stress_value( [ 340, 0, 0, 0 ],
                                    n_steps = 1, load = 0.01 )

    def test_total_strain_energy_value(self):
        '''Assert a correct value of resulting energy in the unit domain
        '''
        self.assert_total_energy_value( 340 * 0.01 / 2.,
                                        n_steps = 3, load = 0.01 )

if __name__ == "__main__":
    import unittest
    #import sys;sys.argv = ['', 'TestMATS2DElastic.test_total_strain_energy_value']
    #import sys;sys.argv = ['', 'TestMATS2DElastic.test_stress_value']
    #import sys;sys.argv = ['', 'TestMATS2DElastic.test_symmetry_shear']
    #import sys;sys.argv = ['', 'TestMATS2DElastic.test_symmetry_tension']
    unittest.main()



import math
import sys
import time

from numpy import array, linspace, pi, arange, sin, cos, ones, frompyfunc, \
    where, hstack
from traits.api import \
    Array, Enum, Float, HasTraits, Int, \
    Property, cached_property, Range, Str, TraitError
class MFnPolar(HasTraits):

    numpoints = Int(80)
    low = Float(0.)
    high = Float(2 * pi)
    # arguments to configure the Polar function to be plotted
    alpha = Range(0., pi, 0., auto_set=False)
    delta_alpha = Range(0., pi / 2, 0., auto_set=False)
    delta_trans = Range(0., pi, 0., auto_set=False)
    phi_residual = Float(0.65)
    phi_quasibrittle = Float(-0.25)
    strech_residual = Range(0., 1, 0., auto_set=False)
    strech_quasibrittle = Range(0., 1, 0., auto_set=False)
    # arguments to configure the plot
    plotrange_min = Float(0., desc='lower bound of the plot range')
    plotrange_max = Float(1., desc='upper bound of the plot range')
    frac_noplot = Range(0., 1, 0.3, desc='fraction of the \
    plot-circle that is not used for display of plot values ', auto_set=False)
    # @todo: put this into the status bar
    info = Str('')

    theta = Array

    def _theta_default(self):
        theta_arr = arange(self.low,
                           self.high,
                           (self.high - self.low) / self.numpoints)
        # add theta=0 to the end of theta-array
        theta_zero = array([0])
        return hstack([theta_arr, theta_zero])

    # function to be plotted (returns radius_value for a given theta_value)
    def radius_fn(self, theta, alpha, delta_alpha, delta_trans,
                  strech_residual, strech_quasibrittle,
                  phi_residual, phi_quasibrittle):
        # 1st quadrant
        if ((theta - alpha) >= 0. and (theta - alpha) <= (pi / 2)):
            theta_tilde = theta - alpha
        # 2nd quadrant
        elif ((theta - alpha) <= pi and (theta - alpha) >= pi / 2):
            theta_tilde = pi - (theta - alpha)
        # 3rd quadrant positive
        elif ((theta - alpha) >= pi and (theta - alpha) <= 3 * pi / 2):
            theta_tilde = theta - alpha - pi
        # 3rd quadrant negative
        elif ((theta - alpha) >= -pi and (theta - alpha) <= -pi / 2):
            theta_tilde = theta - alpha + pi
        # 4th quadrant positive
        elif ((theta - alpha) <= 2 * pi and (theta - alpha) >= 3 * pi / 2):
            theta_tilde = (2 * pi) - (theta - alpha)
        # 4th quadrant negative
        elif ((theta - alpha) <= 0. and (theta - alpha) >= -pi / 2):
            theta_tilde = -(theta - alpha)

        # Definition of function to be plotted in the range of 0 and Pi/2:
        _phi_residual = phi_residual + \
            (1 - phi_residual) * strech_residual
        _phi_quasibrittle = phi_quasibrittle + \
            (1 - phi_quasibrittle) * strech_quasibrittle

        # constant values with linear transition function:
        # (for delta_alpha = 0 the transition function is evaluated)
        if abs(theta_tilde) < delta_alpha:
            radius_fn = _phi_residual
        elif abs(theta_tilde) >= delta_alpha and \
                abs(theta_tilde) < delta_alpha + delta_trans:
            radius_fn = (_phi_residual -
                         ((theta_tilde - delta_alpha) *
                          (_phi_residual - _phi_quasibrittle) / (delta_trans)))
        else:
            radius_fn = _phi_quasibrittle
        return radius_fn

    radius = Property(Array, depends_on='theta,alpha,delta_alpha,\
                                            delta_trans,frac_noplot,\
                                            strech_residual,\
                                            strech_quasibrittle,\
                                            phi_residual,phi_quasibrittle ')

    @cached_property
    def _get_radius(self):
        vradius_fn = frompyfunc(self.radius_fn, 8, 1)
        return array(vradius_fn(self.theta, self.alpha, self.delta_alpha,
                                self.delta_trans, self.strech_residual,
                                self.strech_quasibrittle, self.phi_residual,
                                self.phi_quasibrittle), dtype='float64')

    def __call__(self, theta_value):
        # return a single value for the specified theta_value
        radius_value = self.radius_fn(theta_value, self.alpha,
                                      self.delta_alpha, self.delta_trans,
                                      self.strech_residual,
                                      self.strech_quasibrittle,
                                      self.phi_residual, self.phi_quasibrittle)
        return radius_value

    plot_type = Enum('polar')

#     radiusplot = MFnPolarPlotItem("theta", ["radius", "plotrange_min",
#                                             "plotrange_max", "frac_noplot"],
#                                   type_trait="plot_type",
#
#                                   # Basic axis and label properties
#                                   #                                  show_label=False,
#                                   resizable=False,
#                                   width=260,
#                                   height=260,
#
#                                   # Plot properties
#                                   color="green",
#                                   bgcolor="lightyellow",
#
#                                   # Border, padding properties
#                                   border_visible=True,
#                                   show_label=False,
#                                   border_width=1)

    radius_min = Property(Float, depends_on='current_theta,alpha,delta_alpha,\
                                                delta_trans,strech_residual,\
                                                strech_quasibrittle')

    @cached_property
    def _get_radius_min(self):
        r_min = self.radius.min()
        return r_min

    radius_max = Property(Float, depends_on='current_theta,alpha,delta_alpha,\
                                                delta_trans,strech_residual,\
                                                strech_quasibrittle')

    @cached_property
    def _get_radius_max(self):
        r_max = self.radius.max()
        return r_max

    info = Property(Float, depends_on='current_theta,alpha,delta_alpha,\
                                          delta_trans,strech_residual,\
                                          strech_quasibrittle, plotrange_min, \
                                          plotrange_max')

    @cached_property
    # check for plausibility and return an information or an error message
    def _get_info(self):
        if self.plotrange_min >= self.plotrange_max:
            info = '### Error ###: invalid plot ranges. \
                    plotrange_max must be greater than plotrange_min.'
        elif not (self.plotrange_min <= self.radius_min and
                  self.radius_max <= self.plotrange_max):
            info = '# Note #: Some value of the function are out of \
                    the specified plot ranges!'
        else:
            info = '(valid plot ranges)'
        return info

    current_theta = Float(0.0)

    current_radius = Property(Float, depends_on='current_theta,alpha,\
                                                    delta_alpha,delta_trans,\
                                                    strech_residual,\
                                                    strech_quasibrittle')

    @cached_property
    def _get_current_radius(self):
        return self.__call__(self.current_theta)





from numpy import \
    array, hstack, zeros, vstack, pi as Pi, mgrid, arange, \
    linspace, sin, cos, c_, vstack, zeros, append, concatenate, \
    dot

from traits.api import \
    Str, Callable, Int, List, Bool, Trait
from .mats2D_tensor import map2d_eps_eng_to_mtx
import bmcs_utils.api as bu


class MATS2DRTraceCylinder(bu.Model):

    '''Gather the data for the polar plot

    @todo: define the resolution in the polar direction.

    Limit this model to the set of discretized domain.
    Acquire the microplane geometry from the the model.

    In order to construct the warp, a tensor rank 2 variable must
    be specified in order to make a projection onto the cylinder
    surface. By default, the wapr is called 'eps_app'.  This trait
    attribute may be modified to other derived variables, like
    eps_plastic or eps_dev, whathever the tmodel supports.

    For discrete models, (microplane, the resolution might be given
    in the MATS class itself. Therefore, ask first there.

    @todo: Include the strain field as warp from the material model
    According to the current resolution of the strain space - project
    the strain tensor onto the planes.

    @todo: Add vector field with arrows to show vector magnitudes at individual microplanes.
    '''

    label = Str('MATS2DRTraceCylinder')

    # Switch the warp veriable on and off
    var_warp_on = Bool(True)

    # Variable to be used for the meridian direction
    # (distance between rings)
    var_axis = Str('time')
    var_axis_eval = Callable
    idx_axis = Int(0)

    # Variable to be used for the surface direction
    var_surface = Str('')
    var_surface_eval = Callable
    idx_surface = Int(-1)

    _trace = List()
    _warp_field = List()

    def get_n_t(self):
        '''Get the resolution of the axis.
        This associates every point in the history with a time stamp
        mapped as a z-coordinate into the 3D space.
        '''
        return len(self._trace)

    def get_n_r(self):
        '''
        Get resolution of the polar axis
        '''
        # this is special for the microplane model
        # verify - more general interface to the source material model should be
        # defined
        return 2 * self.var_surface_eval.ts.n_mp

    def __init__(self, **kwargs):
        super(MATS2DRTraceCylinder, self).__init__(**kwargs)
        self._trace = []
        if self.var_warp_on:
            self._warp_field = []

    def bind(self):
        '''
        Locate the evaluators
        '''
        # get the evaluator for the variable
        self.var_surface_eval = self.rmgr.rte_dict[self.var_surface]
        # get the warp_eval callable

    def add_current_values(self, sctx, eps_eng):
        '''
        Invoke the evaluators in the current context for the specified control vector U_k.
        '''
        surface_values = self.var_surface_eval(sctx, eps_eng)
        self._trace.append(surface_values)
        if self.var_warp_on:
            n_r = self.get_n_r()
            alpha_list = array([2 * Pi / n_r * i for i in range(0, n_r)])
            N = array([[cos(alpha), sin(alpha)] for alpha in alpha_list])

            # todo: change this so that it can be applied to a general model.
            #
            # The warp values are in the U_k input given here. It must be mapped to
            # to the strain tensor and then projected onto the planes N
            #
            eps_mtx = map2d_eps_eng_to_mtx(eps_eng)
            warp_values = array([dot(eps_mtx, n) for n in N])

            self._warp_field.append(warp_values)

    #-------------------------------------------------------------------------
    # 3D visualization in MayaVI
    #-------------------------------------------------------------------------
    def xregister_mv_pipelines(self, e):
        '''Register as a source in the pipelane
        '''
        # Remarks[rch]:
        #
        # Pipeline construction
        # ---------------------
        # Maybe this should be done only on demand
        # when the visualization is requested either by the update
        # event triggered here, or by the user. For a batch-like
        # simulation no scenes would be necessary.
        #
        # In that case, the engine could be simply registered and
        # the pipeline construction can be deffered to the explicit
        # request.
        #
        # Further question concerns the multiplicity of the relations.
        # One tracer can construct several sources. The source can be
        # for example the standard array source or parametric surface.
        # There should be no link back to the tracer.
        #
        # Links between tracers and scenes
        # --------------------------------
        # On the other hand, several tracers can contribute to a single
        # scene. Then the tracer explicitly specifies the name
        # of the scene it is contributing to. This type of sharing makes
        # sence if the spatial context of the tracer is the same.
        #
        # The scene management is a separate issue, no general
        # rules can be formulated at this time.
        #
        scene = e.new_scene()
        scene.name = 'Polar domain'

        # Construct the source
        from mayavi.sources.vtk_data_source import VTKDataSource
        from tvtk.api import tvtk

        self._mv_src = VTKDataSource(name='Time-Strain Cylinder',
                                     data=tvtk.PolyData())
        e.add_source(self._mv_src)

        # Construct the warp filter
        if self.var_warp_on:
            from mayavi.filters.api import WarpVector
            e.add_filter(WarpVector())

        # Construct visualization modules
        from mayavi.modules.api import Outline, Surface
        s = Surface()
        e.add_module(Outline())
        e.add_module(s)
        s.module_manager.scalar_lut_manager.show_scalar_bar = True
        s.module_manager.scalar_lut_manager.reverse_lut = True

    def redraw(self):
        '''
        '''
        self.mvp_mgrid_geo.redraw()
        # self.mvp_mgrid_labels.redraw( 'label_scalars' )

    n_t = Int
    n_r = Int

    def _get_points(self):
        '''
        Return an array of points. In order to show the values constant
        on each ring segment, generate two grids so that point values
        from left and right can be stored in order to show the jumps.
        '''
        # to get a reasonable shape of the cylinder define its dimensions
        # The axis values - time-step values must be mapped into
        # the visualization range. Thus, when using axis values, the numbers
        # do not correspond to the real calculated values along the cylinder
        # axis.
        # @todo: Maybe it is possible to adjust this in the axes filter
        #
        T = 10
        R = 2

        self.n_t = n_t = self.get_n_t()
        self.n_r = n_r = self.get_n_r()

        d_r = 2 * Pi / n_r

        alpha_slice1 = slice(0, 2 * Pi - d_r, complex(0, n_r))
        alpha_slice2 = slice(d_r, 2 * Pi, complex(0, n_r))
        axis_slice = slice(0, T, complex(0, n_t))
        grid1 = mgrid[alpha_slice1, axis_slice]
        grid2 = mgrid[alpha_slice2, axis_slice]

        grid1 = array([R * cos(grid1[0]), R * sin(grid1[0]), grid1[1]])
        grid2 = array([R * cos(grid2[0]), R * sin(grid2[0]), grid2[1]])

        points1 = c_[tuple([grid1[i].flatten() for i in range(3)])]
        points2 = c_[tuple([grid2[i].flatten() for i in range(3)])]
        points = vstack([points1, points2])
        return points

    def _get_faces(self):
        '''
        Only return data of n_dims = 2.
        '''
        # Construct  an array of node numbers respecting the grid structure
        # (the nodes are numbered first in t-direction, then in r-direction
        #
        enum_nodes = arange(
            2 * self.n_r * self.n_t).reshape((2 * self.n_r, self.n_t))
        #
        # get the slices extracting all corner nodes with the smallest
        # node number within the element
        #
        enum_nodes_grid1 = arange(
            self.n_r * self.n_t).reshape((self.n_r, self.n_t))
        base_node_list = enum_nodes_grid1[:, 0:-1].flatten()
        #
        # Get the node map within the line
        #
        goffset = self.n_r
        offsets = enum_nodes[[0, 0, goffset, goffset], [0, 1, 1, 0]]
        #
        # The number is determined by putting 1 into inactive dimensions and
        # n_t into the active dimensions.
        #
        n_faces = (self.n_r) * (self.n_t - 1)
        faces = zeros((n_faces, 4), dtype='int_')
        for n_idx, base_node in enumerate(base_node_list):
            faces[n_idx, :] = offsets + base_node
        return faces

    def _get_scalars(self):
        '''Point values to be plotted by  the surface module.
        '''
        f_data = array(self._trace, dtype='float').transpose().flatten()
        f_data = hstack([f_data, f_data])
        return hstack([f_data, f_data])

    def _get_vectors(self):
        '''Vector values that can be used in the warp field.

        The values must be expanded to match the 2-mgrid layout.
        There are two grids in order to be able to represent the
        segment-wise constant values. For a general case,
        this is not necessary. Should the fields be classified
        according to their background mesh? Then,
        '''
        if self.var_warp_on:
            # Convert the list of gathered arrays into
            # a single 2-n_steps array
            #
            # glue the list of deformation vectors together
            #
            f_data = array(self._warp_field, dtype='float')

            # swap axes to have the meridian (time) direction first
            #
            f_data = f_data.swapaxes(0, 1)

            # construct the shifted grid by putting the first
            # line at the end
            #
            g_data = vstack((f_data[1:, :, :], f_data[0:1, 0:, :]))

            # Put the grids together
            #
            f_data = vstack((f_data, g_data))

            # flatten the first two axes to get each point
            # in a single row of a two-dimensional array
            #
            n_points = f_data.shape[0] * f_data.shape[1]
            f_data = f_data.reshape(n_points, 2)

            # Augment with the zero z-vector coordinate
            #
            f_data = hstack([f_data, zeros((n_points, 1), dtype='float')])
            return f_data
        else:
            return []

    def timer_tick(self, e=None):
        pass
        # @todo: unify with redraw
        # self.redraw()

    def clear(self):
        self._trace = []
        if self.var_warp_on:
            self._warp_field = []



'''
Created on Dec 17, 2018

@author: rch
'''
from traits.api import Interface


class IHist(Interface):

    def add_timestep(self, t):
        '''Add time step to history
        '''

    def get_time_idx_arr(self, t):
        '''Get the index corresponding to visual time
        '''

    def get_time_idx(self, t):
        '''Get an index of a time step given the time
        '''



#-------------------------------------------------------------------------
#
# Copyright (c) 2021, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in bmcs/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.bmcs.com/licenses/BSD.txt
#
# Thanks for using BMCS open source!
#

from .i_bcond import \
    IBCond
from ibvpy.mesh.fe_grid_node_slice import FEGridNodeSlice
from ibvpy.tfunction import TimeFunction, TFMonotonic
from ibvpy.mathkit.mfn import MFnLineArray
import numpy as np
import traits.api as tr
import bmcs_utils.api as bu
from .bc_dof import BCDof


@tr.provides(IBCond)
class BCSliceI(bu.Model):
    '''
    Implements the IBC functionality for a constrained dof.
    '''
    name = tr.Str('<unnamed>')

    var = tr.Enum('u', 'f')

    slice = tr.Instance(FEGridNodeSlice)
    link_slice = tr.Instance(FEGridNodeSlice)

    bcdof_list = tr.List(BCDof)

    def reset(self):
        self.bcdof_list = []

    link_coeffs = tr.List(tr.Float)
    '''
    List of dofs that determine the value of the current dof

    If this list is empty, then the current dof is
    prescribed. Otherwise, the dof value is given by the
    linear combination of DOFs in the list (see the example below)

    link_dofs = List( Int )

    Coefficients of the linear combination of DOFs specified in the
    above list.
    '''

    dims = tr.List(tr.Int)

    _link_dims = tr.List(tr.Int)
    link_dims = tr.Property(tr.List(tr.Int))

    def _get_link_dims(self):
        if len(self._link_dims) == 0:
            return self.dims
        else:
            return self._link_dims

    def _set_link_dims(self, link_dims):
        self._link_dims = link_dims

    value = tr.Float

    time_function = tr.Instance(TimeFunction, ())

    def _time_function_default(self):
        return TFMonotonic()

    space_function = tr.Instance(MFnLineArray, ())

    def _space_function_default(self):
        return MFnLineArray(xdata=[0, 1], ydata=[1, 1], extrapolate='diff')

    def is_essential(self):
        return self.var == 'u'

    def is_linked(self):
        return self.link_dofs != []

    def is_constrained(self):
        '''
        Return true if a DOF is either explicitly prescribed or it depends on other DOFS.
        '''
        return self.is_essential() or self.is_linked()

    def is_natural(self):
        return self.var == 'f'

    def setup(self, sctx):
        '''
        Locate the spatial context.f
        '''
        if self.link_slice == None:
            for node_dofs, dof_X in zip(self.slice.dofs,
                                        self.slice.dof_X):
                for dof in node_dofs[self.dims]:
                    self.bcdof_list.append(BCDof(var=self.var,
                                                 dof=dof,
                                                 value=self.value,
                                                 # link_coeffs=self.link_coeffs,
                                                 time_function=self.time_function))
        else:
            # apply the linked slice
            n_link_nodes = len(self.link_slice.dofs.flatten())
            link_dofs = self.link_dofs
            if n_link_nodes == 1:
                #
                link_dof = self.link_slice.dofs.flatten()[0]
                link_coeffs = self.link_coeffs
                for node_dofs, dof_X in zip(self.slice.dofs, self.slice.dof_X):
                    for dof, link_dof, link_coeff in zip(
                            node_dofs[self.dims], link_dofs, link_coeffs):
                        self.bcdof_list.append(BCDof(var=self.var,
                                                     dof=dof,
                                                     link_dofs=[link_dof],
                                                     value=self.value,
                                                     link_coeffs=[link_coeff],
                                                     time_function=self.time_function))
            else:
                for node_dofs, dof_X, node_link_dofs, link_dof_X in \
                    zip(self.slice.dofs, self.slice.dof_X,
                        self.link_slice.dofs, self.link_slice.dof_X):
                    #print('node', node_dofs, node_link_dofs)
                    #print('node[dims]', node_dofs[self.dims],
                          # node_link_dofs[self.link_dims])
                    for dof, link_dof, link_coeff in zip(node_dofs[self.dims],
                                                         node_link_dofs[self.link_dims],
                                                         self.link_coeffs):
                        #print('dof, link, coeff', dof, link_dof, link_coeff)
                        self.bcdof_list.append(BCDof(var=self.var,
                                                     dof=dof,
                                                     link_dofs=[link_dof],
                                                     value=self.value,
                                                     link_coeffs=[
                                                         link_coeff],
                                                     time_function=self.time_function))

    def register(self, K):
        '''Register the boundary condition in the equation system.
        '''
        for bcond in self.bcdof_list:
            bcond.register(K)

    def apply_essential(self, K):

        for bcond in self.bcdof_list:
            bcond.apply_essential(K)

    def apply(self, step_flag, sctx, K, R, t_n, t_n1):

        for bcond in self.bcdof_list:
            bcond.apply(step_flag, sctx, K, R, t_n, t_n1)

    #-------------------------------------------------------------------------
    # Ccnstrained DOFs
    #-------------------------------------------------------------------------

    dofs = tr.Property

    def _get_dofs(self):
        return np.unique(self.slice.dofs[..., self.dims].flatten())

    dof_X = tr.Property

    def _get_dof_X(self):
        return self.slice.dof_X

    n_dof_nodes = tr.Property

    def _get_n_dof_nodes(self):
        sliceshape = self.dofs.shape
        return sliceshape[0] * sliceshape[1]

    #-------------------------------------------------------------------------
    # Link DOFs
    #-------------------------------------------------------------------------
    link_dofs = tr.Property(tr.List)

    def _get_link_dofs(self):
        if self.link_slice != None:
            return np.unique(self.link_slice.dofs[..., self.link_dims].flatten())
        else:
            return []

    link_dof_X = tr.Property

    def _get_link_dof_X(self):
        return self.link_slice.dof_X

    n_link_dof_nodes = tr.Property

    def _get_n_link_dof_nodes(self):
        sliceshape = self.link_dofs.shape
        return sliceshape[0] * sliceshape[1]



#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Sep 8, 2009 by: rch

from math import sin, cos
from numpy import pi as Pi
from traits.api import Range, Float, on_trait_change

from ibvpy.bcond.bc_dof import BCDof
#
# Supply the boundary conditions and construct the response graphs.
#


def get_value_and_coeff(a, alpha):
    '''
    Helper function to construct the kinematic constraint for speficied
    proportion of principle strain components.

    The specification requires that the sum of the projections of e1 and e2
    onto the specified inclined direction alpha is equal to a, i.e. 

    cos(alpha) * e[0] + sin(alpha) * e[1] = a

    The kinematic constraint must be ascribed to a particular DOF and
    can be extended with dependency coefficients. Here we can rewrite 
    the above equation as

    e[0] = - sin(alpha) / cos(alpha) * e[1] + a / cos(alpha) 

    so that the corresponding BCDof specification has the form

    BCDof( var = 'u', # u is used by the time stepper for the kinematic variable
           value = a / cos(alpha),
           dof = 0,
           link_dofs = [1],
           link_coeffs = [-sin(alpha)/cos(alpha) ] )
    '''
    ca = cos(alpha)
    sa = sin(alpha)
    coeff = -(sa / ca)
    value = a / ca
    return value, coeff


class BCDofProportional(BCDof):

    '''Convenience specialization of the BCDof with kinematic link between
    epsilon[0] and epsilon[1] corresponding to the angle - 
    proportion between the two strain components. 

    By changing the level alpha, the boundary conditions
    are  adjusted to the corresponding proportional loading
    of e1 and e2 in the strain space. This allows for elementary
    testing2 of the material point behavior for combined loadings.

    alpha = 0 and 90 means tensile loading in e1 and e2, respectively
    alpha = 45 represents simultaneous tension in e1 and e2
    alpha = 180 and 270 means pure pressure in e1 and e2, respectively

    Note that there is only a single kinematic constraint on strain,
    the remaining strain components are calculated using the 
    equilibrium loop    
    '''
#    alpha_degree = Range( 0., 360., 0.,
#                          label = 'Loading angle',
#                          auto_set = False)

    alpha_rad = Float(0.0,
                      label='Loading angle',
                      enter_set=True,
                      auto_set=False)

    max_strain = Float

    # specialization of the BCDof traits
    var = 'u'
    dof = 0

    def __init__(self, **args):
        super(BCDofProportional, self).__init__(**args)
        self._reset()

    @on_trait_change('alpha_rad,max_strain')
    def _reset(self):
        print('reseting bcond')
        alpha = self.alpha_rad  # Pi * self.alpha_degree / 180
        value, coeff = get_value_and_coeff(self.max_strain, alpha)
        self.value = value
        self.link_dofs = [1]
        self.link_coeffs = [coeff]





from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic
from ibvpy.tmodel.mats2D.mats2D_sdamage import MATS2DScalarDamage
from ibvpy.fets.fets2D.fets2D4q import FETS2D4Q
from ibvpy.fets.fets2D.fets2D4q8u import FETS2D4Q8U
from ibvpy.fets.fets_ls.fets_ls_eval import FETSLSEval #, FETSCracked
from ibvpy.api import\
     BCDofGroup, TStepper as TS, TLoop, TLine, RTDofGraph
from ibvpy.rtrace.rt_domain_list_field import RTraceDomainListField
from ibvpy.mesh.fe_grid import FEGrid
from ibvpy.mesh.fe_refinement_grid import FERefinementGrid

def notched_bended_beam():

    fets_eval_4u      = FETS2D4Q( mats_eval = MATS2DScalarDamage() )
    fets_eval_cracked = FETSLSEval( parent_fets  = fets_eval_4u )

    # Discretization
    fe_domain1 = FEGrid( coord_max = (5.,2.,0.), 
                               shape   = (3,2),
                               fets_eval = fets_eval_4u )

    fe_child_domain = FERefinementGrid( parent_domain = fe_domain1,
                                    fets_eval = fets_eval_cracked,
                                    fine_cell_shape = (1,1) )

    crack_level_set = lambda X: X[0] - 2.5  
    
    fe_child_domain.refine_elem( (1,0), crack_level_set )
    dots = fe_child_domain.new_dots()

    fe_domain  = FEDomainList( subdomains = [ fe_domain1 ] )
    fe_domain_tree = FEDomainTree( domain_list = fe_domain )
    
    ts = TS( dof_resultants = True,
             sdomain = [ fe_domain1, fe_child_domain ],
             bcond_list =  [BCDofGroup(var='u', value = 0., dims = [0,1],
                                       get_dof_method = fe_domain1.get_left_dofs ),
                            BCDofGroup(var='u', value = 0., dims = [0,1],
                                       get_dof_method = fe_domain1.get_right_dofs ),
                            BCDofGroup(var='f', value = -1., dims = [1],
                                       get_dof_method = fe_domain1.get_top_dofs ),
                                       ],
             rtrace_list =  [
#                              RTDofGraph(name = 'Fi,right over u_right (iteration)' ,
#                                   var_y = 'F_int', idx_y = 0,
#                                   var_x = 'U_k', idx_x = 1),
#                        RTraceDomainListField(name = 'Stress' ,
#                             var = 'sig_app', idx = 0, warp = True ),
#                             RTraceDomainField(name = 'Displacement' ,
#                                        var = 'u', idx = 0),
#                                 RTraceDomainField(name = 'N0' ,
#                                              var = 'N_mtx', idx = 0,
#                                              record_on = 'update')
#                          
                    ]             
                )
    
    # Add the time-loop control
    tloop = TLoop( tstepper = ts,
                   tline  = TLine( min = 0.0,  step = 1, max = 1.0 ))
    
    print(tloop.eval())
#    from ibvpy.plugins.ibvpy_app import IBVPyApp
#    ibvpy_app = IBVPyApp( ibv_resource = tloop )
#    ibvpy_app.main()
        
#    print ts.F_int
#    print ts.rtrace_list[0].trace.ydata

if __name__ == '__main__':
    notched_bended_beam()

    


'''

Example demonstrating the interaction between model 
and Visualization sheet.

VizSheet time coincides with the Model time.
The model is required to deliver the time range 
by returning the values of time_low, time_high and time_step. 
By default, the model is expected to return 
the time range as (0,1,10). Updates of the model 
time range should be immediately reflected in the VizSheet. 
An update of the range can be done by the user or by the 
calculation changing the value of t within the model. 
The change of the visual time is reported to the VizXD 
adapters that are requested to change their state.

Provide a dummy model delivering the time range and demonstrate 
the case of interactive steering versus separate calculation 
and browsing steps.

The example should be provided in view.examples.time_control.py

Created on Mar 10, 2017

@author: rch
'''

import time

from traits.api import Instance, List,  HasStrictTraits,\
    Property, cached_property, Bool


from ibvpy.tfunction.tfun_pwl_interactive import TFunPWLInteractive
from ibvpy.bcond import BCDof
from ibvpy.sim.tline import TLine
import numpy as np
from ibvpy.view.window import BMCSWindow
from ibvpy.view.window.bmcs_window import BMCSModel

from .response_tracer import ResponseTracer


class TimeLoop(HasStrictTraits):

    tline = Instance(TLine)

    paused = Bool(False)
    restart = Bool(True)

    def init_loop(self):
        if self.paused:
            self.paused = False
        if self.restart:
            self.tline.val = 0
            self.restart = False

    def eval(self):
        '''this method is just called by the tloop_thread'''

        self.init_loop()
        t_min = self.tline.val
        t_max = self.tline.max
        n_steps = 5
        tarray = np.linspace(t_min, t_max, n_steps)
        for idx, t in enumerate(tarray):
            if self.restart or self.paused:
                break
            time.sleep(1)
            self.tline.val = t


class DemoModel(BMCSModel):
    '''Demo model of the BMCS Window

    Shows how the time control within an application of BMCS
    is realized.

    Run mode
    ========
    The model provides the methods for the Actions 
    Run, Pause, and Continue.

    During these actions, values are registered within
    the response tracers. The model can also send an update
    to the visual object time - so that the response tracers
    are asked to be updated.

    Sliding mode
    ============
    Once the calculation is finished, the slider in the VizSheet
    can be used to browse through the viz-adapters and visualize 
    the response for the current vot.

    Interactive mode
    ================
    Boundary condition can be constructed interactively by
    a boundary condition factory. The factory acts as a Run mode.
    '''
    node_name = 'demo model'

    tree_node_list = List

    def _tree_node_list_default(self):
        return [self.tline, self.rt, self.bc_dof]

    tloop = Property(Instance(TimeLoop))

    @cached_property
    def _get_tloop(self):
        return TimeLoop(tline=self.tline)

    def eval(self):
        self.tloop.eval()

    rt = Instance(ResponseTracer, ())

    bc_dof = Instance(BCDof)

    def _bc_dof_default(self):
        return BCDof(time_function=TFunPWLInteractive())

if __name__ == '__main__':
    model = DemoModel()
    model.run()

    tv = BMCSWindow(model=model)
    model.rt.add_viz2d('time_profile', 'response tracer #1')
    tv.finish_event = True
    tv.configure_traits()




from traits.api import Trait

from ibvpy.tmodel.mats3D.mats3D_cmdm.mats3D_cmdm import MATS3DMicroplaneDamage
from ibvpy.tmodel.mats3D.mats3D_elastic.mats3D_elastic import MATS3DElastic
from ibvpy.tmodel.mats3D.mats3D_microplane import MATS3DMplCSDEEQ
from ibvpy.tmodel.mats3D.mats3D_plastic.mats3D_desmorat import MATS3DDesmorat
from ibvpy.tmodel.mats3D.mats3D_sdamage.vmats3D_sdamage import MATS3DScalarDamage
from ibvpy.tmodel.matsXD.matsXD_explore import MATSXDExplore
from util.traits.either_type import \
    EitherType


class MATS3DExplore(MATSXDExplore):
    '''
    Simulate the loading histories of a material point in 3D space.
    '''

    mats_eval = EitherType(klasses=[MATS3DElastic,
                                    MATS3DDesmorat,
                                    MATS3DMplCSDEEQ,
                                    MATS3DMicroplaneDamage,
                                    MATS3DScalarDamage])

    def _mats_eval_default(self):
        return MATS3DScalarDamage()

    def _mats_eval_changed(self):
        pass



from traits.api import \
    HasStrictTraits, \
    List, \
    on_trait_change, Property, cached_property, \
    Event
import numpy as np


class DomainStateContainer(HasStrictTraits):
    '''Model of the spatial domain - base class approximations and discretizations.

    The XModel represents one subdomain within a spatial domain.
    '''

    def __init__(self, subdomains, *args, **kw):
        super().__init__(*args, **kw)
        self.subdomains = subdomains
        self.serialized_subdomains

    changed_structure = Event

    subdomains = List(domain_changed=True)

    @on_trait_change('changed_structure')
    def _validate_subdomains(self):
        for domain in self.subdomains:
            domain.validate()

    serialized_subdomains = Property(depends_on='subdomains, subdomains_items')

    @cached_property
    def _get_serialized_subdomains(self):
        '''Link the new subdomain at the end of the series.
        '''
        s = np.array(self.subdomains)
        for s1, s2 in zip(s[:-1], s[1:]):
            s1.xmodel.set_next(s2.xmodel)
            s2.xmodel.set_prev(s1.xmodel)
        return self.subdomains

    nonempty_subdomains = Property(depends_on='changed_structure')

    @cached_property
    def _get_nonempty_subdomains(self):
        d_list = []
        for d in self.serialized_subdomains:
            if d.xmodel.n_active_elems > 0:
                d_list.append(d)
        return d_list

    n_dofs = Property

    def _get_n_dofs(self):
        '''Return the total number of dofs in the domain.
        Use the last subdomain's: dof_offset + n_dofs 
        '''
        last_d = self.serialized_subdomains[-1]
        dof_offset = last_d.xmodel.dof_offset
        n_dofs = last_d.xmodel.n_dofs
        return dof_offset + n_dofs

    dof_offset_arr = Property

    def _get_dof_offset_arr(self):
        '''
        Return array of the dof offsets 
        from serialized subdomains
        '''
        a = np.array([domain.xmodel.dof_offset
                      for domain in self.serialized_subdomains])
        return a

    U_var_shape = Property

    def _get_U_var_shape(self):
        return (self.n_dofs,)

    def __iter__(self):
        return iter(self.subdomains)

    def __getitem__(self, idx):
        return self.serialized_subdomains[idx]




from math import sin

from numpy import \
    array, unique, min, max, mgrid, ogrid, c_, repeat, ix_, \
    arange, ones, zeros, multiply, sort, index_exp, frompyfunc, where, \
    zeros_like, sign, delete, ma
from traits.api import \
    HasTraits, List, Array, Property, cached_property, \
    Instance, Trait, Button, on_trait_change, Tuple, \
    Int, Float, provides, WeakRef, Bool, Any, Interface, \
    DelegatesTo, Bool, Callable

from ibvpy.mesh.sdomain import \
    SDomain
# from ibvpy.view.plot3d.mayavi_util.pipelines import \
#     MVPolyData, MVPointLabels, MVStructuredGrid
from ibvpy.mathkit.level_set.level_set import ILevelSetFn, SinLSF, PlaneLSF, ElipseLSF

from .cell_array import CellView, ICellView, CellArray, ICellArraySource
from .cell_grid import CellGrid
from .cell_grid_slice import CellGridSlice


#--------------------------------------------------------------------------
# GeoCellGrid
#--------------------------------------------------------------------------
@provides(ICellArraySource)
class GeoCellGrid(SDomain):

    '''
    Get an array with element node coordinates
    '''
    cell_grid = Instance(CellGrid)

    #-------------------------------------------------------------------------
    # Generation methods for geometry and index maps
    #-------------------------------------------------------------------------
    elem_X_map = Property(depends_on='cell_grid.+')

    @cached_property
    def _get_elem_X_map(self):
        iexp = index_exp[self.cell_grid.cell_node_map]
        return self.cell_grid.point_X_arr[iexp]

    elem_x_map = Property(depends_on='cell_grid.+')

    @cached_property
    def _get_elem_x_map(self):
        iexp = index_exp[self.cell_grid.cell_node_map]
        return self.cell_grid.point_x_arr[iexp]

    def __getitem__(self, idx):
        '''High level access and slicing to the cells within the grid.

        The return value is a tuple with 
        1. array of cell indices
        2. array of nodes for each element
        3. array of coordinates for each node.
        '''
        return GeoGridSlice(geo_grid=self, grid_slice=idx)

    get_cell_point_X_arr = DelegatesTo('cell_grid')
    get_cell_point_x_arr = DelegatesTo('cell_grid')
    get_cell_mvpoints = DelegatesTo('cell_grid')
    cell_node_map = DelegatesTo('cell_grid')
    point_X_grid = DelegatesTo('cell_grid')
    point_x_grid = DelegatesTo('cell_grid')
    point_X_arr = DelegatesTo('cell_grid')
    point_x_arr = DelegatesTo('cell_grid')
    n_dims = DelegatesTo('cell_grid')

    def get_cell_node_labels(self, cell_idx):
        iexp = index_exp[self.cell_grid.cell_node_map[cell_idx]]
        return self.cell_grid.points[iexp]

    #-----------------------------------------------------------------
    # Level set interaction methods
    #-----------------------------------------------------------------

    def get_ls_mask(self, ls_mask_function):
        '''Return a boolean array indicating the masked entries of the level set.
        '''
        X_pnt = self.cell_grid.vertex_X_grid
        vect_fn = frompyfunc(ls_mask_function, self.n_dims, 1)
        ls_mask = vect_fn(*X_pnt)
        return ls_mask

    def _get_transiting_edges_1d(self, ls_function, ls_mask_function=None):

        X_pnt = self.cell_grid.vertex_X_grid

        vect_fn = frompyfunc(ls_function, self.n_dims, 1)

        ls = vect_fn(*X_pnt)

        x_edges = where(ls[:-1] * ls[1:] <= 0)
        return x_edges

    def _get_transiting_edges(self, ls_function, ls_limits=None):

        X_pnt = self.cell_grid.vertex_X_grid

        vect_fn = frompyfunc(ls_function, self.n_dims, 1)
        ls = vect_fn(*X_pnt)

        x_edges = where(ls[:-1, :] * ls[1:, :] <= 0)
        y_edges = where(ls[:, :-1] * ls[:, 1:] <= 0)

        ii, jj = x_edges
        # Get element numbers for each dimension separately
        # for each entry in x_edges
        e_idx = []
        for i, j in zip(ii, jj):
            if j < self.cell_grid.shape[1]:
                e_idx.append([i, j])
            if j > 0:
                e_idx.append([i, j - 1])

        ii, jj = y_edges
        for i, j in zip(ii, jj):
            if i < self.cell_grid.shape[0]:
                e_idx.append([i, j])
            if i > 0:
                e_idx.append([i - 1, j])

        if e_idx == []:
            return e_idx
        else:
            e_exp = array(e_idx, dtype=int).transpose()
            return (e_exp[0, :], e_exp[1, :])

    def get_intersected_cells(self, ls_function, ls_mask_function=None):

        if self.n_dims == 1:
            e_idx = self._get_transiting_edges_1d(
                ls_function, ls_mask_function)
        else:
            e_idx = self._get_transiting_edges(ls_function, ls_mask_function)
        return unique(self.cell_grid.cell_idx_grid[e_idx])

    def get_negative_cells(self, ls_function):
        vect_fn = frompyfunc(ls_function, self.n_dims, 1)
        X_pnt = self.cell_grid.vertex_X_grid
        ls = vect_fn(*X_pnt)

        cutoff_slices = [slice(0, -1) for i in range(self.n_dims)]

        ls = ls[cutoff_slices]

        neg_idx = where(ls < 0)
        negative = unique(self.cell_grid.cell_idx_grid[neg_idx])

        intersected = unique(self.get_intersected_cells(ls_function))

        remaining = list(negative)
        for i in intersected:
            try:
                remaining.remove(i)
            except:
                ValueError

        remaining = array(remaining, dtype=int)
        return remaining

    #-------------------------------------------------------------------------
    # Visualization of level sets
    #-------------------------------------------------------------------------

    def get_mvscalars(self):
        return self.level_set_grid.swapaxes(0, self.cell_grid.n_dims - 1).flatten()

    def _get_ielem_points(self):
        #icells = self.get_elem_intersection()
        icells = self.elem_intersection
        mvpoints = []
        for cell_idx in icells:
            mvpoints += list(self.get_cell_mvpoints(cell_idx))
        return array(mvpoints, dtype='float64')

    def _get_ielem_polys(self):
        #ncells = len( self.get_elem_intersection() )
        ncells = len(self.elem_intersection)
        return arange(ncells * 4).reshape(ncells, 4)

    #-----------------------------------------------------------------
    # Visualization related methods
    #-----------------------------------------------------------------

    refresh_button = Button('Draw')

    @on_trait_change('refresh_button')
    def redraw(self):
        '''Redraw the point grid.
        '''
        self.cell_grid.redraw()

    geo_cell_array = Button

    def _geo_cell_array_fired(self):
        elem_array = self.cell_grid.cell_node_map
        self.show_array = CellArray(data=elem_array,
                                    cell_view=GeoCellView(cell_grid=self))
        self.show_array.current_row = 0
        self.show_array.configure_traits(kind='live')

    #-----------------------------------------------------------------
    # Visualization of level sets related methods
    #-----------------------------------------------------------------

    # mvp_point_grid = Trait(MVStructuredGrid)
    #
    # def _mvp_point_grid_default(self):
    #     return MVStructuredGrid(name='Point grid',
    #                             dims=self.cell_grid._get_mvpoints_grid_shape,
    #                             points=self.cell_grid._get_mvpoints,
    #                             scalars=self.get_mvscalars)
    #
    # mvp_intersect_elems = Trait(MVPolyData)
    #
    # def _mvp_intersect_elems_default(self):
    #     return MVPolyData(name='Intersected elements',
    #                       points=self._get_ielem_points,
    #                       polys=self._get_ielem_polys)

    ls_refresh_button = Button('Draw Levelset')

    @on_trait_change('ls_refresh_button')
    def ls_redraw(self):
        '''Redraw the point grid.
        '''
        self.mvp_point_grid.redraw()
        self.mvp_intersect_elems.redraw()

class GeoGridSlice(CellGridSlice):

    geo_grid = WeakRef(GeoCellGrid)

    def __init__(self, geo_grid, **args):
        self.geo_grid = geo_grid
        super(GeoGridSlice, self).__init__(**args)

    cell_grid = Property(depends_on='geo_grid.+changed_structure')

    @cached_property
    def _get_cell_grid(self):
        return self.geo_grid.cell_grid

    point_X_arr = Property

    def _get_point_X_arr(self):
        idx1, idx2 = self.idx_tuple
        return self.geo_grid.elem_X_map[ix_(self.elems, self.cell_grid.grid_cell[idx2])]

    point_x_arr = Property

    def _get_point_x_arr(self):
        idx1, idx2 = self.idx_tuple
        return self.geo_grid.elem_x_map[ix_(self.elems, self.cell_grid.grid_cell[idx2])]

#-- Tabular Adapter Definition -------------------------------------------

#-----------------------------------------------------------------------
# View a single cell instance
#-----------------------------------------------------------------------


class GeoCellView(CellView):

    '''View a single cell instance.
    '''
    # implements(ICellView)

    cell_X_arr = Array

    cell_x_arr = Array

    def set_cell_traits(self):
        self.cell_X_arr = self.cell_grid.get_cell_point_X_arr(self.cell_idx)
        self.cell_x_arr = self.cell_grid.get_cell_point_x_arr(self.cell_idx)

    #-----------------------------------------------------------------------
    # Visualization
    #-----------------------------------------------------------------------
    draw_cell = Bool(False)


    def redraw(self):
        if self.draw_cell:
            self.mvp_cell_node_labels.redraw()

    #-----------------------------------------------------------------------
    # Private methods
    #-----------------------------------------------------------------------
    def _get_cell_mvpoints(self):
        return self.cell_grid.get_cell_mvpoints(self.cell_idx)

    def _get_cell_node_labels(self):
        return self.cell_grid.get_cell_node_labels(self.cell_idx)

    def _get_cell_lines(self):
        return self.cell_grid.grid_cell_spec.cell_lines



'''
Created on Mar 24, 2011

@author: jakub
'''

from numpy import \
    array, zeros,  dot, hstack, \
    identity
from scipy.linalg import \
    inv
from traits.api import \
    Instance, Int, Trait, Dict, \
    DelegatesTo, Property
from ibvpy.fets.fets_eval import FETSEval, RTraceEvalElemFieldVar

#-------------------------------------------------------------------------
# FETS2D4Q - 4 nodes iso-parametric quadrilateral element (2D, linear, Lagrange family)
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
# Element Information:
#-------------------------------------------------------------------------
#
# Here an isoparametric element formulation is applied.
# The implemented shape functions are derived based on the
# ordering of the nodes of the parent element defined in
# '_node_coord_map' (see below)
#
#-------------------------------------------------------------------------


class FETS2DTF(FETSEval):

    debug_on = True
    parent_fets = Instance(FETSEval)

    # Dimensional mapping
    dim_slice = slice(0, 2)
    dof_r = DelegatesTo('parent_fets')
    geo_r = DelegatesTo('parent_fets')

    get_dNr_geo_mtx = DelegatesTo('parent_fets')

    get_N_geo_mtx = DelegatesTo('parent_fets')

    n_nodal_dofs = Int(4)

    n_e_dofs = Property(Int)

    def _get_n_e_dofs(self):
        return self.parent_fets.n_e_dofs * 2

    ngp_r = Property

    def _get_ngp_r(self):
        return self.parent_fets.ngp_r

    ngp_s = Property

    def _get_ngp_s(self):
        return self.parent_fets.ngp_s

    vtk_cell_types = Property

    def _get_vtk_cell_types(self):
        return self.parent_fets.vtk_cell_types

    vtk_cells = Property

    def _get_vtk_cells(self):
        return self.parent_fets.vtk_cells

    vtk_r = Property

    def _get_vtk_r(self):
        return self.parent_fets.vtk_r

    #---------------------------------------------------------------------
    # Method required to represent the element geometry
    #---------------------------------------------------------------------

    #---------------------------------------------------------------------
    # Method delivering the shape functions for the field variables and their derivatives
    #---------------------------------------------------------------------
    def get_N_mtx(self, r_pnt):
        '''
        Returns the matrix of the shape functions used for the field approximation
        containing zero entries. The number of rows corresponds to the number of nodal
        dofs. The matrix is evaluated for the specified local coordinate r.
        '''
        n_nodes = self.n_e_dofs / self.n_nodal_dofs
        p_N_mtx = self.parent_fets.get_N_mtx(r_pnt)[0, ::2]
        I_mtx = identity(self.n_nodal_dofs, float)
        N_mtx_list = [I_mtx * p_N_mtx[i] for i in range(0, n_nodes)]
        N_mtx = hstack(N_mtx_list)
        return N_mtx

    def get_dNr_mtx(self, r_pnt):
        '''
        Return the derivatives of the shape functions
        '''
        return self.parent_fets.get_dNr_mtx(r_pnt)

    def get_B_mtx(self, r_pnt, X_mtx):
        J_mtx = self.get_J_mtx(r_pnt, X_mtx)
        dNr_mtx = self.get_dNr_mtx(r_pnt)
        dNx_mtx = dot(inv(J_mtx), dNr_mtx)
        N_mtx = self.get_N_mtx(r_pnt)
        N_mtx_red = N_mtx[:2] - N_mtx[2:]
        Bx_mtx = zeros((8, self.n_e_dofs), dtype='float64')
        Bx_mtx[[0, 2], ::4] = dNx_mtx[[0, 1]]
        Bx_mtx[[3, 5], 2::4] = dNx_mtx[[0, 1]]
        Bx_mtx[[1, 2], 1::4] = dNx_mtx[[1, 0]]
        Bx_mtx[[4, 5], 3::4] = dNx_mtx[[1, 0]]
        Bx_mtx[6:] = N_mtx_red
        return Bx_mtx

    def get_eps_m(self, sctx, u):
        X_mtx = sctx.X
        r_pnt = sctx.loc
        B_mtx = self.get_B_mtx(r_pnt, X_mtx)
        eps = dot(B_mtx, u)
        return array([[eps[0], eps[2]], [eps[2], eps[1]]])

    def get_eps_f(self, sctx, u):
        X_mtx = sctx.X
        r_pnt = sctx.loc
        B_mtx = self.get_B_mtx(r_pnt, X_mtx)
        eps = dot(B_mtx, u)
        return array([[eps[3], eps[5]], [eps[5], eps[4]]])

    def get_eps_b(self, sctx, u):
        X_mtx = sctx.X
        r_pnt = sctx.loc
        B_mtx = self.get_B_mtx(r_pnt, X_mtx)
        eps = dot(B_mtx, u)
        return array([[eps[6], 0.], [0., eps[7]]])

    def get_u_m(self, sctx, u):
        N_mtx = self.get_N_mtx(sctx.loc)
        return dot(N_mtx, u)[:2]

    def get_u_f(self, sctx, u):
        N_mtx = self.get_N_mtx(sctx.loc)
        return dot(N_mtx, u)[2:]

    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        '''
        RTraceEval dictionary with standard field variables.
        '''
        rte_dict = self._debug_rte_dict()
        for key, v_eval in list(self.mats_eval.rte_dict.items()):

            # add the eval into the loop.
            #
            rte_dict[key] = RTraceEvalElemFieldVar(name=key,
                                                   u_mapping=self.map_eps,
                                                   eval=v_eval)

        rte_dict.update({'eps_m': RTraceEvalElemFieldVar(eval=self.get_eps_m),
                         'u_m': RTraceEvalElemFieldVar(eval=self.get_u_m),
                         'eps_f': RTraceEvalElemFieldVar(eval=self.get_eps_f),
                         'u_f': RTraceEvalElemFieldVar(eval=self.get_u_f),
                         'eps_b': RTraceEvalElemFieldVar(eval=self.get_eps_b)})

        return rte_dict


#----------------------- example --------------------

def example_with_new_domain():
    from ibvpy.api import \
        TStepper as TS, RTraceDomainListField, TLoop, TLine

    from ibvpy.tmodel.mats2D5.mats2D5_bond.mats2D_bond import MATS2D5Bond
    from ibvpy.api import BCDofGroup
    from ibvpy.fets.fets2D.fets2D4q import FETS2D4Q
    fets_eval = FETS2DTF(parent_fets=FETS2D4Q(),
                         mats_eval=MATS2D5Bond(E_m=30, nu_m=0.2,
                                               E_f=10, nu_f=0.1,
                                               G=10.))

    from ibvpy.mesh.fe_grid import FEGrid
    from mathkit.mfn import MFnLineArray

    # Discretization
    fe_grid = FEGrid(coord_max=(10., 4., 0.),
                     n_elems=(10, 3),
                     fets_eval=fets_eval)

    mf = MFnLineArray(  # xdata = arange(10),
        ydata=array([0, 1, 2, 3]))

    tstepper = TS(sdomain=fe_grid,
                  bcond_list=[BCDofGroup(var='u', value=0., dims=[0, 1],
                                         get_dof_method=fe_grid.get_left_dofs),
                              #                                   BCDofGroup( var='u', value = 0., dims = [1],
                              # get_dof_method = fe_grid.get_bottom_dofs ),
                              BCDofGroup(var='u', value=.005, dims=[0],
                                         time_function=mf.get_value,
                                         get_dof_method=fe_grid.get_right_dofs)],
                  rtrace_list=[
                      #                     RTDofGraph(name = 'Fi,right over u_right (iteration)' ,
                      #                               var_y = 'F_int', idx_y = right_dof,
                      #                               var_x = 'U_k', idx_x = right_dof,
                      #                               record_on = 'update'),
                      #                         RTraceDomainListField(name = 'Stress' ,
                      #                         var = 'sig_app', idx = 0,
                      #                         #position = 'int_pnts',
                      #                         record_on = 'update'),
                      #                     RTraceDomainListField(name = 'Damage' ,
                      #                                    var = 'omega', idx = 0,
                      #                                    record_on = 'update',
                      #                                    warp = True),
                      RTraceDomainListField(name='Displ matrix',
                                            var='u_m', idx=0,
                                            record_on='update',
                                            warp=True),
                      RTraceDomainListField(name='Displ reinf',
                                            var='u_f', idx=0,
                                            record_on='update',
                                            warp=True),

                      #                    RTraceDomainListField(name = 'N0' ,
                      #                                      var = 'N_mtx', idx = 0,
                      # record_on = 'update')
                  ]
                  )

    # Add the time-loop control
    #global tloop
    tloop = TLoop(tstepper=tstepper, KMAX=300, tolerance=1e-4,
                  tline=TLine(min=0.0, step=1.0, max=1.0))

    #import cProfile
    #cProfile.run('tloop.eval()', 'tloop_prof' )
    print(tloop.eval())
    #import pstats
    #p = pstats.Stats('tloop_prof')
    # p.strip_dirs()
    # print 'cumulative'
    # p.sort_stats('cumulative').print_stats(20)
    # print 'time'
    # p.sort_stats('time').print_stats(20)

    # Put the whole thing into the simulation-framework to map the
    # individual pieces of definition into the user interface.
    #
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    app = IBVPyApp(ibv_resource=tloop)
    app.main()

if __name__ == '__main__':
    example_with_new_domain()



'''
Created on May 18, 2009

@author: jakub
'''
from ibvpy.fets.fets2D.fets2D4q import FETS2D4Q
from ibvpy.fets.fets3D.fets3D8h20u import FETS3D8H20U
from ibvpy.fets.fets_eval import FETSEval
from ibvpy.mesh.fe_grid import FEGrid

if __name__ == '__main__':

    example = '2D'

    if example == '1D':
        fets_sample = FETSEval( dof_r = [[-1], [0], [1]],
                               geo_r = [[-1], [1]],
                               n_nodal_dofs = 1 )

        fe_domain = FEGrid( coord_max = ( 2., ),
                            shape = ( 2, ),
        #                              inactive_elems = [3],
                            fets_eval = fets_sample )
        print("n_e ", fe_domain.shape)
        #second element,right node 
        fe_slice = fe_domain[1, -1]
        #second element
        #fe_slice = fe_domain[1] 
    elif example == '2D':
        fets_sample = FETS2D4Q()

        fe_domain = FEGrid( coord_max = ( 3., 2. ),
                            shape = ( 3, 2 ),
        #                              inactive_elems = [3],
                            fets_eval = fets_sample )
        #first element right top node 
        fe_slice = fe_domain[2, 1, -1, -1]

    elif example == '3D':
        fets_sample = FETS3D8H20U()

        fe_domain = FEGrid( coord_max = ( 3., 2., 4 ),
                            shape = ( 3, 2, 4 ),
        #                              inactive_elems = [3],
                            fets_eval = fets_sample )

        # first element, right front edge                  
        fe_slice = fe_domain[0, 0, 0, -1, :, -1]

    print('elems')
    print(fe_slice.elems)
    print('dof_nodes')
    print(fe_slice.dof_nodes)
    print('dofs')
    print(fe_slice.dofs)
    print('dof coords')
    print(fe_slice.dof_X)
    print('geo_nodes')
    print(fe_slice.geo_nodes)
    print('geo_r')
    print(fe_slice.geo_X)
    print('points? what are they for?')
